from bigflow._version import __version__
from bigflow.commons import public

import bigflow.workflow
import bigflow.configuration


__all__ = [
    'Workflow',
    'Job',
    'JobContext',
    'Definition',
    'Config',
]


@public(class_alias=True)
class Definition(bigflow.workflow.Definition): ...


@public(class_alias=True)
class Job(bigflow.workflow.Job): ...


@public(class_alias=True)
class JobContext(bigflow.workflow.JobContext): ...


@public(class_alias=True)
class Workflow(bigflow.workflow.Workflow): ...


@public(class_alias=True)
class Config(bigflow.configuration.Config): ...


# proactively try to initialize bigflow-specific logging
# it is used to configure logging on pyspark/beam/etc workers
try:
    from bigflow.log import maybe_init_logging_from_env
except ImportError:
    pass  # logging is not installed?
else:
    maybe_init_logging_from_env()
    del maybe_init_logging_from_env



from ..configuration import Config
from .interface import Dataset
from .interactive import InteractiveDatasetManager


class DatasetConfig:
    def __init__(self,
                 env: str,
                 project_id: str,
                 dataset_name: str = 'None',
                 internal_tables: list = None,
                 external_tables: dict = None,
                 properties: dict = None,
                 is_master: bool = True,
                 is_default: bool = True):
       all_properties = (properties or {}).copy()
       all_properties['project_id'] = project_id
       all_properties['dataset_name'] = dataset_name
       all_properties['internal_tables'] = internal_tables or []
       all_properties['external_tables'] = external_tables or {}

       self.delegate = Config(name=env, properties=all_properties, is_master=is_master, is_default=is_default)

    def add_configuration(self,
                          env: str,
                          project_id: str,
                          dataset_name: str = None,
                          internal_tables: list = None,
                          external_tables: dict = None,
                          properties: dict = None,
                          is_default: bool = False):

        all_properties = (properties or {}).copy()

        all_properties['project_id'] = project_id

        if dataset_name:
            all_properties['dataset_name'] = dataset_name

        if internal_tables:
            all_properties['internal_tables'] = internal_tables

        if external_tables:
            all_properties['external_tables'] = external_tables

        self.delegate.add_configuration(env, all_properties, is_default=is_default)
        return self

    def create_dataset_manager(self, env: str = None) -> Dataset:
        return InteractiveDatasetManager(
            project_id=self.resolve_project_id(env),
            dataset_name=self.resolve_dataset_name(env),
            internal_tables=self.resolve_internal_tables(env),
            external_tables=self.resolve_external_tables(env),
            extras=self.resolve_extra_properties(env))

    def resolve_extra_properties(self, env: str = None):
        return {k: v for (k, v) in self.resolve(env).items() if self._is_extra_property(k)}

    def pretty_print(self, env_name: str = None):
        return self.delegate.pretty_print(env_name)

    def __str__(self):
        return str(self.delegate)

    def resolve(self, env_name: str = None) -> dict :
        return self.delegate.resolve(env_name)

    def resolve_property(self, property_name: str, env: str = None):
        return self.delegate.resolve_property(property_name, env)

    def resolve_project_id(self, env: str = None) -> str:
        return self.resolve_property('project_id', env)

    def resolve_dataset_name(self, env: str = None) -> str:
        return self.resolve_property('dataset_name', env)

    def resolve_internal_tables(self, env: str = None) -> str:
        return self.resolve_property('internal_tables', env)

    def resolve_external_tables(self, env: str = None) -> str:
        return self.resolve_property('external_tables', env)

    def _is_extra_property(self, property_name) -> bool:
        return property_name not in ['project_id','dataset_name','internal_tables','external_tables', 'env']


import os
import unittest

import tblib  # ensure installed

import bigflow.testing.isolate

from . import nonpure


class TestIsolateMixinTestCase(unittest.TestCase):

    class SubTest(unittest.TestCase):
        failed = False

        @classmethod
        def setUpClass(cls):
            cls.pid = os.getpid()

        def test_fail(self):
            type(self).failed = True
            self.fail('fail-ok')

        def test_ok(self):
            self.assertNotEqual(self.pid, os.getpid())

        def test_error(self):
            raise RuntimeError('fail-error')

        @unittest.skip('test-skip')
        def test_skip(self):
            self.fail()

        @unittest.expectedFailure
        def test_expected_failure_but(self):
            pass

        @unittest.expectedFailure
        def test_expected_failure(self):
            self.fail("expected")

    class ForkTest(bigflow.testing.ForkIsolateMixin, SubTest):
        pass

    class SpawnTest(bigflow.testing.SpawnIsolateMixin, SubTest):
        pass

    def check_test_mixin_subtest(self, test_class):
        # when
        result = self._run_test(test_class)

        # then
        self.assertEqual(6, result.testsRun, "All tests should run")

        # then
        self.assertEqual(1, len(result.failures))
        self.assertRegex(result.failures[0][0].id(), r".*\.test_fail", "Test 'test_fail' sould fail")
        self.assertIsNotNone(result.failures[0][1], "Traceback is attached")
        self.assertFalse(test_class.failed, "State is not propogated")

        # then]
        self.assertEqual(1, len(result.errors))
        self.assertRegex(result.errors[0][0].id(), ".*\.test_error", "Test 'test_error' sould raise exception")
        self.assertIsNotNone(result.errors[0][1], "Traceback is attached")

        # then
        self.assertEqual(1, len(result.skipped), "Single skipped test")
        self.assertEqual(1, len(result.unexpectedSuccesses), "Single unexpected success")
        self.assertEqual(1, len(result.expectedFailures), "Single expected failure")

    def _run_test(self, test_class):
        devnull = open(os.devnull, 'wt')
        self.addCleanup(devnull.close)
        runner = unittest.TextTestRunner(stream=devnull)
        suite = unittest.defaultTestLoader.loadTestsFromTestCase(test_class)
        return runner.run(suite)

    @unittest.skip
    def test_fork_mixin(self):
        self.check_test_mixin_subtest(self.ForkTest)

    def test_spawn_mixin(self):
        self.check_test_mixin_subtest(self.SpawnTest)


class ForkReloadModulesTest(
    bigflow.testing.ForkIsolateMixin,
    unittest.TestCase,
):
    def setUpParent(self):
        super().setUpParent()
        self.const = nonpure.CONST

    def test_test(self):
        from . import nonpure
        self.assertEqual(self.const, nonpure.CONST, "Module should not be reloaded")


class SpawnReloadModulesTest(
    bigflow.testing.SpawnIsolateMixin,
    unittest.TestCase,
):
    def setUpParent(self):
        super().setUpParent()
        self.const = nonpure.CONST

    def test_test(self):
        from . import nonpure
        self.assertNotEqual(self.const, nonpure.CONST, "Module should be reloaded")


class _IsolateDisabledBase(
    bigflow.testing.isolate._IsolatedProcessMixin,
    unittest.TestCase,
):
    isolate = False
    _state = None

    @classmethod
    def _change_state(cls, from_, to_):
        if cls._state != from_:
            raise AssertionError(f"invalid state, expected {from_} but got {cls._state}")
        cls._state = to_

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._change_state(None, 'setupClass')

    def setUpParent(self):
        super().setUpParent()
        self._change_state('setupClass', 'setupParent')

    def setUp(self):
        super().setUp()
        self._change_state('setupParent', 'setup')

    def test_test(self):
        #self.assertEqual(self.thepid, os.getpid())
        self._change_state('setup', 'run')

    def tearDown(self):
        self._change_state('run', 'teardown')
        super().tearDown()

    def tearDownParent(self):
        self._change_state('teardown', 'teardownParent')
        super().tearDownParent()

    @classmethod
    def tearDownClass(cls):
        cls._change_state('teardownParent', None)
        super().tearDownClass()


class ForkDisabledTest(
    _IsolateDisabledBase,
    bigflow.testing.ForkIsolateMixin,
    unittest.TestCase,
):
    pass


class SpawnDisabledTest(
    _IsolateDisabledBase,
    bigflow.testing.SpawnIsolateMixin,
    unittest.TestCase,
):
    pass


import io
import os
import logging
import pprint
import typing as T

from bigflow.commons import public


logger = logging.getLogger(__name__)



def current_env():
    """Returns current env name (specified via 'bigflow --config' option)"""
    return os.environ.get('bf_env')


@public()
class Config:

    def __init__(self,
        name: str,
        properties: T.Dict[str, str],
        is_master: bool = True,
        is_default: bool = True,
    ):
        self.master_properties = properties if is_master else {}
        self.default_env_name = None
        self.configs = {}
        self.environment_variables_prefix = 'bf_'

        self.add_configuration(name, properties, is_default)

    def __str__(self):
        return "".join(map(self.pretty_print, self.configs.keys())).rstrip("\n")

    def resolve_property(self, property_name: str, env_name: str = None):
        try:
            return self.resolve(env_name)[property_name]
        except KeyError:
            raise ValueError(
                f"Failed to load property '{property_name}' from config, "
                f"also there is no '{self.environment_variables_prefix}{property_name}' env variable.")

    def pretty_print(self, env_name: str = None):
        s = io.StringIO()
        pp = pprint.PrettyPrinter(indent=4, stream=s)
        _, env_name = self._get_env_config(env_name)

        s.write(env_name)
        s.write(" config:\n")
        pp.pprint(self.resolve(env_name))

        return s.getvalue()

    def _capture_osenv_properties(self):
        prefix = self.environment_variables_prefix
        prefix_len = len(prefix)
        return {
            k[prefix_len:]: v
            for k, v in os.environ.items()
            if k.startswith(prefix)
        }

    def resolve(self, env_name: str = None) -> dict:
        env_config, env_name = self._get_env_config(env_name)

        properties_with_placeholders = dict(env_config)
        for k, v in self._capture_osenv_properties().items():
            if properties_with_placeholders.get(k, None) is None:
                properties_with_placeholders[k] = v

        for k, v in properties_with_placeholders.items():
            if v is None:
                raise ValueError(
                    f"Failed to load property '{k}' from OS environment, "
                    f"no such env variable: '{self.environment_variables_prefix}{k}'.")

        return {
            key: self._resolve_placeholders(value, properties_with_placeholders)
            for key, value in properties_with_placeholders.items()
        }

    def add_configuration(self, name: str, properties: dict, is_default: bool = False):
        props = {}
        props.update(self.master_properties)
        props.update(properties)

        assert 'env' not in properties or properties['env'] == name
        props['env'] = name

        self.configs[name] = props
        self._update_default_env_name(name, is_default)
        return self

    def _update_default_env_name(self, name: str, is_default: bool):
        if not is_default:
            return
        if self.default_env_name:
            raise ValueError(f"default env is already set to '{self.default_env_name}', you can set only one default env")
        self.default_env_name = name

    def _get_env_config(self, name: str) -> T.Tuple[dict, str]:
        explicit_env_name = name or os.environ.get(f'{self.environment_variables_prefix}env')
        if not explicit_env_name:
            if not self.default_env_name:
                raise ValueError("No explicit env name is given and no default env is defined, can't resolve properties.")
            return self.configs[self.default_env_name], self.default_env_name

        try:
            return self.configs[explicit_env_name], explicit_env_name
        except KeyError:
                raise ValueError(f"no such config name '{explicit_env_name}'")

    def _resolve_placeholders(self, value, variables: dict):
        if isinstance(value, str):
            modified_value = value
            for k, v in variables.items():
                if isinstance(v, str) and v != value:
                    modified_value = modified_value.replace("{%s}" % k, v)
            return modified_value
        else:
            return value


@public()
class DeploymentConfig(Config):
    def __init__(self,
                 name: str,
                 properties: dict,
                 is_master: bool = True,
                 is_default: bool = True,
                 environment_variables_prefix: str = None):
        super().__init__(
            name=name,
            properties=properties,
            is_master=is_master, is_default=is_default)
        self.environment_variables_prefix = environment_variables_prefix or 'bf_'



"""Integrates bigflow build toolchain with 'distutils'"""

import os
import sys
from typing import Dict, Optional, Union
import setuptools
import logging
import pickle

import distutils.cmd
import distutils.command.sdist
import distutils.dist
import distutils.filelist
import distutils.log as dlog

from pathlib import Path
from datetime import datetime

import bigflow.cli
import bigflow.resources
import bigflow.dagbuilder
import bigflow.version
import bigflow.build.pip
import bigflow.build.dev
import bigflow.build.operate
import bigflow.build.spec as spec


logger = logging.getLogger(__name__)

SETUP_VALIDATION_MESSAGE = 'BigFlow setup is valid.'


class BigflowDistribution(distutils.dist.Distribution):
    """Customized Distribution for bigflow projects. Add custom commands, allow access to bigflow project spec."""

    def __init__(self, attrs: Optional[Dict]=None):
        self.bigflow_project_spec = None
        attrs = dict(attrs or {})
        cmdclass = attrs.setdefault('cmdclass', {})
        cmdclass['build_project'] = build_project
        cmdclass['sdist'] = sdist
        super().__init__(attrs)

    def get_command_class(self, command: str):
        cls = super().get_command_class(command)
        if command.startswith("bdist"):
            cls = _hook_pregenerate_sdist(cls)
        return cls


def _hook_pregenerate_sdist(command_cls):
    """
    Wraps existing distutils.Command class.
    Runs 'sdist' and copy result into 'build/bf-project.tar.gz'
    """

    def run(self: distutils.cmd.Command):
        distribution: BigflowDistribution = self.distribution

        # build sdist package & copy into /build
        sdist = self.get_finalized_command('sdist')
        sdist.ensure_finalized()
        sdist.formats = ["tar"]  # overwrite
        sdist.run()
        sdist_tarball = sdist.get_archive_files()

        if len(sdist_tarball) > 1:
            self.warn("ingnored 'sdist' results", sdist_tarball[1:])
        self.mkpath("build")
        self.copy_file(sdist_tarball[0], "build/bf-project.tar")

        # generate patched pyproject.toml inside /build
        pyproject_toml = Path("build", "pyproject.toml")
        if Path("pyproject.toml").exists():
            self.copy_file("pyproject.toml", pyproject_toml)
        spec.add_spec_to_pyproject_toml(pyproject_toml, distribution.bigflow_project_spec)

        return command_cls.run(self)

    return type(
        command_cls.__name__,
        (command_cls,),
        {'run': run},
    )


class sdist(distutils.command.sdist.sdist):
    """Customized `sdist` command.

    - Add custom items to MANIFEST
    - Write project spec into 'pyproject.toml'
      (freeze calculated properties, like version etc)
    - Generate shim 'setup.py' when it is missing
    """

    distribution: BigflowDistribution

    def make_release_tree(self, base_dir, files):
        super().make_release_tree(base_dir, files)
        self._generate_files(base_dir)

    def add_defaults(self):
        super().add_defaults()
        self._add_defaults_bigflow()

    def _generate_files(self, base_dir):

        # append generated project specs to pyproject.toml
        p = Path(base_dir) / "pyproject.toml"

        dlog.info("add full bigflow project info to %s", p)
        if p.exists():  # break hard links
            content = p.read_bytes()
            p.unlink()
            p.write_bytes(content)
        spec.add_spec_to_pyproject_toml(p, self.distribution.bigflow_project_spec)

        # generate shim 'setup.py'
        if not p.exists():
            p.write_text("""from bigflow.build import setup; setup()""")

    def _add_defaults_bigflow(self):
        self.filelist.extend(
            filter(os.path.exists, [
                "setup.py",
                "pyproject.toml",
                "deployment_config.py",
                "requirements.in",
                "requirements.txt",
                "Dockerfile",
                "resources/requirements.txt",
                "resources/requirements.in",
            ]))


# Deprecated: there is no more need for this command since 1.3
# It exists however for backward compatability.
class build_project(distutils.cmd.Command):

    distribution: BigflowDistribution

    description = 'BigFlow project build.'
    user_options = [
        ('build-dags', None, 'Builds the DAG files.'),
        ('build-package', None, 'Builds the whl package.'),
        ('build-image', None, 'Builds the Docker image.'),
        ('start-time=', None, 'DAGs start time -- given in local timezone, for example: 2020-06-27 15:00:00'),
        ('workflow=', None, 'The workflow that you want to build DAG for.'),
        ('validate-project-setup', None, 'If used, echoes a message that can be used by the CLI to determine if the setup is working.'),
    ]

    def initialize_options(self) -> None:
        self.start_time = datetime.now().strftime("%Y-%m-%d %H:00:00")
        self.build_dags = False
        self.build_package = False
        self.build_image = False
        self.workflow = None
        self.validate_project_setup = False

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        print("Note: don't use `setup.py` directly, instead use `bigflow build` command-line tool")
        if self.validate_project_setup:
            print(SETUP_VALIDATION_MESSAGE)
            return
        prj = self.distribution.bigflow_project_spec
        if self.build_package:
            bigflow.build.operate.build_package(prj)
        elif self.build_dags:
            bigflow.build.operate.build_dags(prj, self.start_time, self.workflow)
        elif self.build_image:
            bigflow.build.operate.build_image(prj)
        else:
            bigflow.build.operate.build_project(prj, self.start_time, self.workflow)


def projectspec_to_setuppy_kwargs(p: spec.BigflowProjectSpec):
    attrs = {
        'distclass': BigflowDistribution,
        'bigflow_project_spec': p,
        'name': p.name,
        'version': p.version,
        'packages': p.packages,
        'install_requires': p.requries,
        'data_files': [
            ('resources', list(bigflow.resources.find_all_resources(p.project_dir / p.resources_dir))),
            (f"bigflow__project/{p.name}", ["build/bf-project.tar", "build/pyproject.toml"]),
            *(p.data_files or []),
        ],
        'script_name': "setup.py",
        **p.metainfo,
        **p.setuptools,
    }
    # filter None values
    attrs = {k: v for k, v in attrs.items() if v is not None}
    return attrs


def run_setup_command(prj: spec.BigflowProjectSpec, command, options=None):
    """Execute distutils command in the scope of same python process."""

    attrs = projectspec_to_setuppy_kwargs(prj)
    logger.debug("Create tmp Distribution with attrs %r", attrs)
    dist = BigflowDistribution(attrs)

    if options:
        logger.debug("Update command options with %s", options)
        dist.get_option_dict(command).update(options)

    cmd_obj = dist.get_command_obj(command)
    logger.debug("Command object is %s", cmd_obj)
    cmd_obj.ensure_finalized()

    logger.info("Run command %s with options %s", command, options)
    cmd_obj.run()


def _maybe_dump_setup_params(params):
    if len(sys.argv) == 3 and sys.argv[1] == bigflow.build.dev.DUMP_PARAMS_SETUPPY_CMDARG:
        with open(sys.argv[2], 'w+b') as out:
            pickle.dump(params, out)
        sys.exit(0)


def setup(project_dir=None, **kwargs):
    _maybe_dump_setup_params(kwargs)

    logging.basicConfig(level=logging.INFO, format="%(message)s")
    logger.info("Run bigflow.build.setup...")

    project_dir = Path(project_dir or Path.cwd())
    prj = spec.read_project_spec_nosetuppy(project_dir=project_dir, **kwargs)
    setuppy_kwargs = projectspec_to_setuppy_kwargs(prj)

    logger.debug("setuptools.setup(**%r)", setuppy_kwargs)
    return setuptools.setup(
        **setuppy_kwargs,
    )


import textwrap
import unittest
from unittest import mock

from bigflow.version import bump_minor, release

from test import mixins


class GetVersionE2E(
    mixins.TempCwdMixin,
    mixins.SubprocessMixin,
    mixins.BigflowInPythonPathMixin,
    unittest.TestCase,
):

    def get_version(self):
        return self.subprocess_run([
            "python", "-c", textwrap.dedent("""
                import bigflow.version
                print(bigflow.version.get_version())
            """)
            ],
            text=True,
        ).stdout.strip()

    def test_should_version_based_on_git_tags(self):
        # then
        self.assertRegex(self.get_version(), r"^0\+BROKEN$", "No git repo")

        # when
        self.subprocess_run("git init")
        # then
        self.assertRegex(self.get_version(), r"^0\+BROKEN$", "Empty git repo")

        # when
        (self.cwd / "file1").touch()
        self.subprocess_run("git add file1")
        self.subprocess_run("git commit -m message")
        # then
        self.assertRegex(self.get_version(), r"^0\+g.+$", "Single commit, no tags")

        # when
        (self.cwd / "file1").write_text("changed")
        # then
        self.assertRegex(self.get_version(), r"^0\+g.+\.t.+$$", "Single commit, no tags, dirty")

        # when
        self.subprocess_run("git add file1")
        self.subprocess_run("git commit -m message")
        self.subprocess_run("git tag 0.2.0")
        # then
        self.assertRegex(self.get_version(), r"^0.2.0$", "Single tag, exact match")

        # when
        (self.cwd / "file1").write_text("changed2")
        # then
        self.assertRegex(self.get_version(), r"^0.2.0.dev0\+g.{8,}\.t.+$", "Single tag, dirty")

        # when
        self.subprocess_run("git add file1")
        self.subprocess_run("git commit -m message")
        # then
        self.assertRegex(self.get_version(), r"^0.2.0.dev1\+g.{8,}", "No exact tag matched")

        # when
        (self.cwd / "file1").write_text("change4")
        # then
        self.assertRegex(self.get_version(), r"^0.2.0.dev1\+g.{8,}\.t.+$", "No exact tag matched, dirty")


class ReleaseTestCase(unittest.TestCase):

    @mock.patch('bigflow.version.push_tag')
    @mock.patch('bigflow.version.get_tag')
    def test_should_push_bumped_tag(self, get_tag_mock, push_tag_mock):
        # given
        get_tag_mock.return_value = None

        # when
        release('fake_pem_path')

        # then
        push_tag_mock.assert_called_with("0.1", 'fake_pem_path')

        # given
        get_tag_mock.return_value = '0.2'

        # when
        release('fake_pem_path')

        # then
        push_tag_mock.assert_called_with('0.3', 'fake_pem_path')


class BumpMinorTestCase(unittest.TestCase):

    def test_should_bump_minor_(self):
        for fr, to in [
            # full version
            ("1.0.0", "1.1"),
            ("0.1.0", "0.2"),
            ("0.1.1", "0.2"),
            ("0.0.1", "0.1"),
            ("0.1.dev1", "0.2"),

            # only major
            ("0", "0.1"),
            ("12", "12.1"),

            # preserve prefix
            ("v10.1", "v10.2"),
        ]:
            self.assertEqual(bump_minor(fr), to)

    def test_should_raise_value_error_for_invalid_version_schema(self):
        # given
        invalid_version = 'some-garbage'

        # then
        with self.assertRaises(ValueError):
            # when
            bump_minor(invalid_version)


import functools
import hashlib
import logging

from inspect import getfullargspec

import typing
from pathlib import Path

from google.api_core.exceptions import BadRequest

import bigflow

from .job import Job
from .job import DEFAULT_RETRY_COUNT
from .job import DEFAULT_RETRY_PAUSE_SEC
from .dataset_manager import DEFAULT_LOCATION
from .interface import Dataset, DEFAULT_RUNTIME

logger = logging.getLogger(__name__)

DEFAULT_OPERATION_NAME = '__auto'
DEFAULT_PEEK_LIMIT = 1000

INLINE_COMPONENT_DATASET_ALIAS = '_inline_component_dataset'


def log_syntax_error(method):
    logger.debug("Wrap %s with @log_syntax_error", method)

    @functools.wraps(method)
    def decorated(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except BadRequest as e:
            if 'Syntax error' in e.message:
                logger.error(e.message)
            else:
                raise e

    return decorated


def interactive_component(**dependencies):
    def decorator(standard_component):
        logger.debug("Wrap %s with @interactive_component, deps %s", standard_component, dependencies)
        return InteractiveComponent(standard_component,
                                    {dep_name: dep.config for dep_name, dep in dependencies.items()})
    return decorator


class InteractiveDatasetManager(Dataset):
    """Let's you run operations on a dataset, without the need of creating a component."""

    def __init__(self,
                 project_id,
                 dataset_name,
                 internal_tables=None,
                 external_tables=None,
                 credentials=None,
                 extras=None,
                 location=DEFAULT_LOCATION):
        self.config = DatasetConfigInternal(
            project_id=project_id,
            dataset_name=dataset_name,
            internal_tables=internal_tables,
            external_tables=external_tables,
            credentials=credentials,
            extras=extras,
            location=location)
        logger.debug("Create InteractiveDatasetManager, config %s", self.config._as_dict())

    def write_truncate(self, table_name, sql, partitioned=True):
        method = 'write_truncate'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name=table_name, sql=sql),
            method,
            table_name,
            sql,
            partitioned=partitioned,
            operation_name=DEFAULT_OPERATION_NAME)

    def write_append(self, table_name, sql, partitioned=True):
        method = 'write_append'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name=table_name, sql=sql),
            method,
            table_name,
            sql,
            partitioned=partitioned,
            operation_name=DEFAULT_OPERATION_NAME)

    def write_tmp(self, table_name, sql):
        method = 'write_tmp'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name=table_name, sql=sql),
            method,
            table_name,
            sql,
            operation_name=DEFAULT_OPERATION_NAME)

    def collect(self, sql):
        method = 'collect'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name='', sql=sql),
            method,
            sql,
            operation_name=DEFAULT_OPERATION_NAME)

    def collect_list(self, sql: str, record_as_dict: bool = False):
        method = 'collect_list'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name='', sql=sql),
            method,
            sql,
            record_as_dict=record_as_dict,
            operation_name=DEFAULT_OPERATION_NAME)

    def dry_run(self, sql):
        method = 'dry_run'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name='', sql=sql),
            method,
            sql,
            operation_name=DEFAULT_OPERATION_NAME)

    def create_table(self, create_query):
        method = 'create_table'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name='', sql=create_query),
            method,
            create_query,
            operation_name=DEFAULT_OPERATION_NAME)

    def load_table_from_dataframe(self, table_name, df, partitioned=True):
        method = 'load_table_from_dataframe'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name=table_name, sql=''),
            method,
            table_name=table_name,
            df=df,
            partitioned=partitioned,
            operation_name=DEFAULT_OPERATION_NAME)

    def create_table_from_schema(
            self,
            table_name: str,
            schema: typing.Union[typing.List[dict], Path, None] = None,
            table=None):
        method = 'create_table_from_schema'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name=table_name, sql=''),
            method,
            table_name=table_name,
            schema=schema,
            table=table)

    def insert(
            self,
            table_name: str,
            records: typing.Union[typing.List[dict], Path],
            partitioned: bool = True):
        method = 'insert'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name=table_name, sql=''),
            method,
            table_name=table_name,
            records=records,
            partitioned=partitioned,
            operation_name=DEFAULT_OPERATION_NAME)

    def delete_dataset(self):
        method = 'delete_dataset'
        return self._tmp_interactive_component_factory(
            generate_component_name(method=method, table_name=self.config.dataset_name, sql=''),
            method,
            operation_name=DEFAULT_OPERATION_NAME)

    def _tmp_interactive_component_factory(self, component_name, method, *args, **kwargs):
        logger.debug("Build tmp interactive component, name=%s, method=%s", component_name, method)

        @interactive_component(_inline_component_dataset=self)
        def tmp_component(_inline_component_dataset):
            return getattr(_inline_component_dataset, method)(*args, **kwargs)

        tmp_component._standard_component.__name__ = component_name
        return tmp_component


def generate_component_name(method, table_name, sql):
    """
    >>> generate_component_name('write_truncate', 'some_table', 'select * from another_table')
    'write_truncate_some_table_5ada1f6e843dc3c517d5eedbae557fbbdf98d6d5047272f092e7b89455826722'
    """
    component_id = hashlib.sha256()
    component_id.update(sql.encode('utf-8'))
    component_id = component_id.hexdigest()
    res = '{}_{}_{}'.format(method, table_name, component_id)
    logger.debug("Generated component name is %s", res)
    return res


class InteractiveComponent(object):
    """Let's you run the component for the specific runtime
     and peek the operation results as the pandas.DataFrame."""

    def __init__(self, standard_component, dependency_config):
        self._standard_component = standard_component
        self._dependency_config = dependency_config

    def to_job(self,
               id=None,
               retry_count=DEFAULT_RETRY_COUNT,
               retry_pause_sec=DEFAULT_RETRY_PAUSE_SEC,
               dependencies_override=None):
        _, component_callable = decorate_component_dependencies_with_operation_level_dataset_manager(self._standard_component)

        dependencies_override = dependencies_override or {}
        dependency_config = self._dependency_config.copy()
        dependency_config.update({dataset_alias: dataset.config for dataset_alias, dataset in dependencies_override.items()})

        logger.info("Convert interactive component to job, id=%s, component %s, dependencies %s", id, component_callable, dependency_config)
        return Job(
            component=component_callable,
            id=id,
            retry_count=retry_count,
            retry_pause_sec=retry_pause_sec,
            **dependency_config)

    @log_syntax_error
    def run(self, runtime=DEFAULT_RUNTIME, operation_name=None):
        _, component_callable = decorate_component_dependencies_with_operation_level_dataset_manager(
            self._standard_component, operation_name=operation_name)
        job = Job(component_callable, **self._dependency_config)
        logger.info("Run interactive component, id=%s, component %s", job.id, job.component)
        return job.execute(bigflow.JobContext.make(runtime=runtime))

    @log_syntax_error
    def peek(self, runtime, operation_name=DEFAULT_OPERATION_NAME, limit=DEFAULT_PEEK_LIMIT):
        """Returns the result of the specified operation in the form of the pandas.DataFrame, without really running the
        operation and affecting the table."""

        if runtime is None:
            raise ValueError(f"'runtime' can't be None")
        if operation_name is None:
            raise ValueError(f"'operation_name' can't be None")
        if limit is None:
            raise ValueError(f"'limit' can't be None")

        results_container, component_callable = decorate_component_dependencies_with_operation_level_dataset_manager(
            self._standard_component, operation_name=operation_name, peek=True, peek_limit=limit)

        job = Job(component_callable, **self._dependency_config)
        job.execute(bigflow.JobContext.make(runtime=runtime))

        try:
            return results_container[0]
        except IndexError:
            if operation_name != DEFAULT_OPERATION_NAME:
                raise ValueError("Operation '{}' not found".format(operation_name))
            else:
                raise ValueError("You haven't specified operation_name.".format(operation_name))

    def __call__(self, **kwargs):
        _, component_callable = decorate_component_dependencies_with_operation_level_dataset_manager(self._standard_component)
        return component_callable(**kwargs)


def decorate_component_dependencies_with_operation_level_dataset_manager(
        standard_component,
        operation_name=None,
        peek=None,
        peek_limit=None):

    logger.debug(
        "Decorate component dependencies with operation level dataset manager: component %s, operation %s, peek %s/%s",
        standard_component, operation_name, peek, peek_limit,
    )

    operation_settings = {'operation_name': operation_name, 'peek': peek, 'peek_limit': peek_limit}
    results_container = []

    def component_callable(**kwargs):
        operation_level_dataset_managers = {k: OperationLevelDatasetManager(v, **operation_settings)
                                            for k, v in kwargs.items()}

        component_return_value = standard_component(**operation_level_dataset_managers)

        for operation_level_dataset_manager in operation_level_dataset_managers.values():
            if operation_level_dataset_manager._result:
                results_container.extend(operation_level_dataset_manager._result)

        return component_return_value

    component_callable_with_proper_signature = "def reworked_function({signature}):\n    return original_func({kwargs})\n".format(
        signature=','.join([arg for arg in getfullargspec(standard_component).args]),
        kwargs=','.join(['{arg}={arg}'.format(arg=arg) for arg in getfullargspec(standard_component).args]))

    logger.debug("Compile dynamic python code\n%s", component_callable_with_proper_signature)
    component_callable_with_proper_signature_code = compile(component_callable_with_proper_signature, "fakesource",
                                                            "exec")
    fake_locals = {}
    eval(component_callable_with_proper_signature_code, {'original_func': component_callable}, fake_locals)

    component_callable = fake_locals['reworked_function']
    component_callable.__name__ = operation_name or standard_component.__name__

    return results_container, component_callable


class OperationLevelDatasetManager(Dataset):
    """
    Let's you run specified operation or peek a result of a specified operation.
    """

    def __init__(self, dataset_manager, peek=False, operation_name=None, peek_limit=DEFAULT_PEEK_LIMIT):
        self._dataset_manager = dataset_manager
        self._peek = peek
        self._operation_name = operation_name
        self._peek_limit = peek_limit
        self._results_container = []

    def write_truncate(self, table_name, sql, partitioned=True, custom_run_datetime=None, operation_name=None):
        return self._run_operation(
            operation_name=operation_name,
            method=self._dataset_manager.write_truncate,
            sql=sql,
            table_name=table_name,
            partitioned=partitioned,
            custom_run_datetime=custom_run_datetime)

    def write_append(self, table_name, sql, partitioned=True, custom_run_datetime=None, operation_name=None):
        return self._run_operation(
            operation_name=operation_name,
            method=self._dataset_manager.write_append,
            sql=sql,
            table_name=table_name,
            partitioned=partitioned,
            custom_run_datetime=custom_run_datetime)

    def write_tmp(self, table_name, sql, custom_run_datetime=None, operation_name=None):
        return self._run_operation(
            operation_name=operation_name,
            method=self._dataset_manager.write_tmp,
            sql=sql,
            table_name=table_name,
            custom_run_datetime=custom_run_datetime)

    def collect(self, sql, custom_run_datetime=None, operation_name=None):
        return self._run_operation(
            operation_name=operation_name,
            method=self._dataset_manager.collect,
            sql=sql,
            custom_run_datetime=custom_run_datetime)

    def collect_list(
            self,
            sql: str,
            custom_run_datetime: typing.Optional[str] = None,
            record_as_dict: bool = False,
            operation_name=None):
        return self._run_operation(
            operation_name=operation_name,
            method=self._dataset_manager.collect_list,
            sql=sql,
            custom_run_datetime=custom_run_datetime,
            record_as_dict=record_as_dict)

    def dry_run(self, sql, custom_run_datetime=None, operation_name=None):
        return self._run_operation(
            operation_name=operation_name,
            method=self._dataset_manager.dry_run,
            sql=sql,
            custom_run_datetime=custom_run_datetime)

    def create_table(self, create_query, operation_name=None):
        if self._should_run_operation(operation_name):
            return self._results_container, self._dataset_manager.create_table(create_query=create_query)

    def load_table_from_dataframe(self, table_name, df, partitioned=True, custom_run_datetime=None, operation_name=None):
        if self._should_peek_operation_results(operation_name):
            return df
        elif self._should_run_operation(operation_name):
            return self._dataset_manager.load_table_from_dataframe(
                table_name=table_name,
                df=df,
                custom_run_datetime=custom_run_datetime,
                partitioned=partitioned)

    def create_table_from_schema(
            self,
            table_name: str,
            schema: typing.Union[typing.List[dict], Path, None] = None,
            table=None,
            operation_name=None):
        if self._should_run_operation(operation_name):
            return self._results_container, self._dataset_manager.create_table_from_schema(
                table_name=table_name,
                schema=schema,
                table=table)

    def insert(
            self,
            table_name: str,
            records: typing.Union[typing.List[dict], Path],
            partitioned: bool = True,
            operation_name=None):
        if self._should_peek_operation_results(operation_name):
            return records
        elif self._should_run_operation(operation_name):
            return self._results_container, self._dataset_manager.insert(
                table_name=table_name,
                records=records,
                partitioned=partitioned)

    def delete_dataset(self, operation_name=None):
        if self._should_run_operation(operation_name):
            return self._results_container, self._dataset_manager.remove_dataset()

    @property
    def dt(self):
        return self._dataset_manager.runtime_str

    @property
    def extras(self):
        return self._dataset_manager.extras

    @property
    def _result(self):
        return self._results_container

    @property
    def client(self):
        return self._dataset_manager.client

    @property
    def project_id(self):
        return self._dataset_manager.project_id

    @property
    def dataset_name(self):
        return self._dataset_manager.dataset_name

    @property
    def internal_tables(self):
        return self._dataset_manager.internal_tables

    @property
    def external_tables(self):
        return self._dataset_manager.external_tables

    def _collect_select_result_to_pandas(self, sql):
        sql = sql if 'limit' in sql.lower() else sql + '\nLIMIT {}'.format(str(self._peek_limit))
        return self._dataset_manager.collect(sql)

    def _run_operation(self, operation_name, method, sql, *args, **kwargs):
        if self._should_peek_operation_results(operation_name):
            result = self._collect_select_result_to_pandas(sql)
            self._results_container.append(result)
            return result
        elif self._should_run_operation(operation_name):
            return method(*args, sql=sql, **kwargs)
        else:
            import pandas as pd
            return pd.DataFrame()

    def _should_peek_operation_results(self, operation_name):
        return self._operation_name == operation_name and self._peek is not None

    def _should_run_operation(self, operation_name):
        return self._operation_name == operation_name or self._operation_name is None


class DatasetConfigInternal(object):

    def __init__(self,
                 project_id,
                 dataset_name,
                 internal_tables=None,
                 external_tables=None,
                 credentials=None,
                 extras=None,
                 location=DEFAULT_LOCATION):
        self.project_id = project_id
        self.dataset_name = dataset_name
        self.internal_tables = internal_tables or []
        self.external_tables = external_tables or {}
        self.credentials = credentials or None
        self.extras = extras or {}
        self.location = location

    def _as_dict(self):
        return {
            'project_id': self.project_id,
            'dataset_name': self.dataset_name,
            'internal_tables': self.internal_tables,
            'external_tables': self.external_tables,
            'credentials': self.credentials,
            'extras': self.extras,
            'location': self.location
        }


def sensor(table_alias, where_clause, ds=None):

    def sensor_function(ds):
        result = ds.collect('''
        SELECT count(*) > 0 as table_ready
        FROM `{%(table_alias)s}`
        WHERE %(where_clause)s
        ''' % {
            'table_alias': table_alias,
            'where_clause': where_clause
        })

        if not result.iloc[0]['table_ready']:
            raise ValueError('{} is not ready'.format(table_alias))

    sensor_function.__name__ = 'wait_for_{}'.format(table_alias)

    return sensor_function if ds is None else interactive_component(ds=ds)(sensor_function)


def add_label(table_name, labels, ds=None):

    def add_label_function(ds):
        table = ds.client.get_table("{}.{}.{}".format(ds.project_id, ds.dataset_name, table_name))
        table.labels = labels
        return ds.client.update_table(table, ["labels"])

    if ds is None:
        return add_label_function
    else:
        return interactive_component(ds=ds)(add_label_function)


from unittest import TestCase
from unittest import main
import pandas as pd
import mock
from bigflow.bigquery.interactive import sensor as sensor_component

YESTERDAY_WHERE_CLAUSE = "DATE(%(partitioning_column)s) = DATE(TIMESTAMP_ADD(TIMESTAMP('{dt} UTC'), INTERVAL -24 HOUR))"


def return_table_not_ready(*args, **kwargs):
    return pd.DataFrame([{'table_ready': 0}])


def return_table_ready(*args, **kwargs):
    return pd.DataFrame([{'table_ready': 1}])


class SensorComponentTestCase(TestCase):

    def test_should_generate_component_name_based_on_table_alias(self):
        # given
        dataset = mock.Mock()
        sensor = sensor_component('some_table', YESTERDAY_WHERE_CLAUSE, ds=dataset)

        # expect
        self.assertEqual(sensor._standard_component.__name__, 'wait_for_some_table')

    def test_should_raise_error_when_query_result_is_empty(self):
        # given
        dataset = mock.Mock()
        dataset.collect.side_effect = return_table_not_ready
        sensor = sensor_component('some_table', YESTERDAY_WHERE_CLAUSE, ds=dataset)

        # expect
        with self.assertRaises(ValueError):
            sensor(ds=dataset)

    def test_should_generate_query_based_on_where_clause_and_table_alias(self):
        # given
        dataset = mock.Mock()
        sensor = sensor_component('some_table', YESTERDAY_WHERE_CLAUSE % {
            'partitioning_column': 'partition'
        }, ds=dataset)
        dataset.collect.side_effect = return_table_ready

        # when
        sensor(ds=dataset)

        # then
        dataset.collect.assert_called_once_with(sql='''
        SELECT count(*) > 0 as table_ready
        FROM `{some_table}`
        WHERE DATE(partition) = DATE(TIMESTAMP_ADD(TIMESTAMP('{dt} UTC'), INTERVAL -24 HOUR))
        ''', custom_run_datetime=None)


if __name__ == '__main__':
    main()


import re
import logging
import subprocess
import typing as T
import tempfile
import pathlib

import bigflow.commons


logger = logging.getLogger(__name__)


def run_process(
        *args: T.Union[str, T.List],
        verbose: bool = False,
        **kwargs) -> str:
    return bigflow.commons.run_process(*args, verbose=verbose, **kwargs)


def get_version() -> str:
    if not _is_git_available():
        logger.warning("No git repo is available")
        return f"0+BROKEN"

    dirty = _generate_dirty_suffix()

    if not dirty:
        # try direct tag match
        try:
            tag = run_process(["git", "describe", "--exact-match", "--dirty=+dirty", "--tags"])
        except subprocess.SubprocessError as e:
            logger.debug("No direct git tag match: %s", e)
        else:
            m = re.fullmatch(r"""(?x)
                \s*\D*           # strip any prefix
                (?P<tag>\d.*)    # version, begins with a number
                \s*
            """, tag)
            if not m:
                raise ValueError("Invalid tag", tag)
            return m.group('tag')

    # try generic 'git describe' (fail when there are no tags)
    try:
        version = run_process(["git", "describe", "--abbrev=8", "--long", "--tags"])
    except subprocess.SubprocessError as e:
        logger.debug("No long git describtion: %s", e)
    else:
        # replace last '-' with '+' to be PEP440 compliant, add 'dev' to distance
        logger.debug("Parse git describe output %r", version)
        m = re.fullmatch(r"""(?x)
            \s*\D*                    # strip any prefix
            (?P<tag>\d.*?)            # version, begins with a number
            -
            (?P<dev>\d+)              # distance to the nearest tag
            -
            (?P<ghash>g[a-f\d]{8,})  # commit git-hash
            \s*
        """, version)
        return "{tag}.dev{dev}+{ghash}{dirty}".format(dirty=dirty, **m.groupdict())

    # no tags? mayb just githash will work
    try:
        ghash = run_process(["git", "rev-parse", "HEAD"])[:12]
    except subprocess.SubprocessError as e:
        logger.debug("No githash available: %s", e)
    else:
        return f"0+g{ghash}{dirty}"

    logger.error("Can't detect project version based on git")
    return f"0+BROKEN{dirty}"


def _is_git_available() -> bool:
    try:
        run_process(["git", "rev-parse", "--is-inside-work-tree"])
    except subprocess.SubprocessError:
        return False
    else:
        return True


def _generate_dirty_suffix() -> str:
    try:
        dirty = bool(run_process(["git", "diff", "--shortstat", "HEAD"]).strip())
    except subprocess.SubprocessError as e:
        logger.error("Unable to run git diff: %s", e)
        return ""

    if dirty:
        tree = _get_workdir_treehash()
        return f".t{tree[:8]}"
    else:
        return ""


def _get_workdir_treehash() -> str:
    """Copy git index, add all changed files and returns 'tree-hash'"""

    git_root =  run_process(["git", "rev-parse", "--show-toplevel"]).rstrip("\n")
    indexf = pathlib.Path(git_root) / ".git" / "index"

    with tempfile.NamedTemporaryFile(buffering=0) as tf:
        env = {"GIT_INDEX_FILE": tf.name}
        tf.write(indexf.read_bytes())

        run_process(["git", "add", "-u"], env_add=env)
        return run_process(["git", "write-tree"], env_add=env)


def get_tag() -> T.Optional[str]:
    """Return the last tag for the git repository reachable from HEAD."""
    tags = run_process(["git", "tag", "--sort=version:refname", "--merged"]).splitlines()
    if not tags:
        logger.warning("No existing git tags found, use '0.0'")
        return None
    logger.debug("Found %d git tags, latest is %s", len(tags), tags[-1])
    return tags[-1]


def release(identity_file: T.Optional[str] = None) -> None:
    latest_tag = get_tag()
    if not latest_tag:
        logger.warning("No existing git tags found, use '0.0'")
        latest_tag = "0.0"
    new_tag = bump_minor(latest_tag)
    push_tag(new_tag, identity_file)


def push_tag(tag: str, identity_file: T.Optional[str] = None) -> None:
    logger.info("Create git tag %s", tag)
    run_process(["git", "tag", tag])

    if identity_file is not None:
        logger.info("Pushing using the specified identity_file: %s", identity_file)
        env_add = {"GIT_SSH_COMMAND": f"ssh -i {identity_file} -o IdentitiesOnly=yes"}
    else:
        env_add = None

    logger.info("Push tag %s to origin", tag)
    run_process(["git", "push", "origin", tag], env_add=env_add)


def bump_minor(version: str) -> str:
    """Increment minor version part, leaves rest of the version string unchanged"""
    m = re.fullmatch(r"""(?x)
        (?P<prefix>\D*)?     # optional prefix
        (?P<major>\d+)       # major part
        (\.(?P<minor>\d+))?  # minor part
        (?P<suffix>.*)       # unsed rest
    """, version)
    if not m:
        raise ValueError("Invalid version string", version)

    prefix = m.group('prefix') or ""
    major = m.group('major')
    minor = m.group('minor') or "0"
    minor2 = str(int(minor) + 1)

    return f"{prefix}{major}.{minor2}"



import os
import shutil
import datetime
import subprocess

from datetime import timedelta
from pathlib import Path
from unittest import TestCase, mock

from test import mixins

import bigflow.build.spec as spec
from bigflow.commons import run_process
from bigflow.cli import walk_module_files
from bigflow.build.dist import SETUP_VALIDATION_MESSAGE

from bigflow.commons import (
    get_docker_image_id,
    build_docker_image_tag,
)
from bigflow.build.operate import (
    clear_image_leftovers,
    clear_dags_leftovers,
    clear_package_leftovers,
    build_image,
)

PROJECT_NAME = 'main_package'
DOCKER_REPOSITORY = 'test_docker_repository'
TEST_PROJECT_PATH = None


def mkdir(dir_path: Path):
    if not os.path.isdir(dir_path):
        os.mkdir(dir_path)


def rmdir(dir_path: Path):
    if dir_path.exists() and dir_path.is_dir():
        shutil.rmtree(dir_path)


def create_image_leftovers(test_project_dir_path=None):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH

    mkdir(test_project_dir_path / '.image')
    (test_project_dir_path / '.image' / 'leftover').touch()


def create_package_leftovers(
        test_project_dir_path=None,
        project_name: str = PROJECT_NAME):

    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    mkdir(test_project_dir_path / 'build')
    (test_project_dir_path / 'build' / 'leftover').touch()
    mkdir(test_project_dir_path / 'dist')
    (test_project_dir_path / 'dist' / 'leftover').touch()
    mkdir(test_project_dir_path / f'{project_name}.egg-info')
    (test_project_dir_path / f'{project_name}.egg-info' / 'leftover').touch()


def create_dags_leftovers(test_project_dir_path=None):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    mkdir(test_project_dir_path / '.dags')
    (test_project_dir_path / '.dags' / 'leftover').touch()


def dags_leftovers_exist(test_project_dir_path=None):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    return os.path.exists(test_project_dir_path / '.dags' / 'leftover')


def package_leftovers_exist(
        test_project_dir_path=None,
        project_name: str = PROJECT_NAME):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    return os.path.exists(test_project_dir_path / 'build' / 'leftover') and\
           os.path.exists(test_project_dir_path / 'dist' / 'leftover') and\
           os.path.exists(test_project_dir_path / f'{project_name}.egg-info' / 'leftover')


def image_leftovers_exist(test_project_dir_path=None):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    return os.path.exists(test_project_dir_path / '.image' / 'leftover')


def dir_not_empty(dir_path: Path):
    return len(os.listdir(dir_path)) != 0


def deployment_config_copied(test_project_dir_path=None):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    return (test_project_dir_path / '.image' / 'deployment_config.py').exists()


def python_package_built(test_project_dir_path=None):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    return (test_project_dir_path / 'dist' / 'main_package-0.1.0-py3-none-any.whl').exists()


def is_test_run(test_project_dir_path=None):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    return (test_project_dir_path / 'build' / 'junit-reports').exists()


def dags_built(test_project_dir_path: Path, expected_workflow_count: int):
    if not (test_project_dir_path / '.dags').exists():
        return False
    wff = [
        workflow_name for workflow_dir, workflow_name in
        walk_module_files(test_project_dir_path / '.dags')
    ]
    workflows_count = sum(
        1 for workflow_dir, workflow_name in
        walk_module_files(test_project_dir_path / '.dags')
        if 'workflow' in workflow_name
    )
    return workflows_count == expected_workflow_count


def docker_image_as_file_built(test_project_dir_path=None):
    test_project_dir_path = test_project_dir_path or TEST_PROJECT_PATH
    return (test_project_dir_path / '.image' / 'image-0.1.0.tar').exists()


def docker_image_built_in_registry(docker_repository: str, version: str):
    return get_docker_image_id(build_docker_image_tag(docker_repository, version))


def dags_contain(test_project_dir_path: Path, substring: str):
    for module_dir, module_file_name in walk_module_files(test_project_dir_path / '.dags'):
        with open(os.path.join(module_dir, module_file_name), 'r') as f:
            content = f.read()
            if substring not in content:
                return False
    return True


class SetupTestCase(
    mixins.PrototypedDirMixin,
    mixins.SubprocessMixin,
    mixins.BigflowInPythonPathMixin,
    TestCase,
):
    proto_dir = "bf-projects/example_project"

    def setUp(self) -> None:
        super().setUp()
        self.prj = spec.parse_project_spec(
            name=PROJECT_NAME,
            project_dir=self.cwd,
            docker_repository=DOCKER_REPOSITORY,
            requries=[],
        )

        global TEST_PROJECT_PATH
        TEST_PROJECT_PATH = self.cwd

    def run_build(self, cmd):
        return self.subprocess_run(cmd, text=True).stdout


class BuildProjectE2E(SetupTestCase):

    def test_should_build_project_artifacts(self):
        # when
        create_package_leftovers()
        create_image_leftovers()
        create_dags_leftovers()
        self.run_build('python setup.py build_project')

        # then
        self.assertTrue(python_package_built())
        self.assertTrue(is_test_run())
        self.assertTrue(dags_built(TEST_PROJECT_PATH, 2))

        self.assertTrue(docker_image_as_file_built())
        self.assertFalse(docker_image_built_in_registry(DOCKER_REPOSITORY, '0.1.0'))
        self.assertTrue(deployment_config_copied())
        self.assertTrue((TEST_PROJECT_PATH / "resources" / "requirements.txt").exists())

        # and
        self.assertFalse(package_leftovers_exist())
        self.assertFalse(image_leftovers_exist())
        self.assertFalse(dags_leftovers_exist())

    def test_should_validate_project_setup(self):
        # expected
        self.assertTrue(SETUP_VALIDATION_MESSAGE in self.run_build('python setup.py build_project --validate-project-setup'))
        self.assertTrue(SETUP_VALIDATION_MESSAGE in self.run_build('python setup.py build_project --build-dags --validate-project-setup'))
        self.assertTrue(SETUP_VALIDATION_MESSAGE in self.run_build('python setup.py build_project --build-image --validate-project-setup'))
        self.assertTrue(SETUP_VALIDATION_MESSAGE in self.run_build('python setup.py build_project --build-package --validate-project-setup'))


class BuildPackageCommandE2E(SetupTestCase):
    def test_should_execute_build_package_command(self):
        # given
        create_package_leftovers(self.prj.project_dir)
        clear_image_leftovers(self.prj)
        clear_dags_leftovers(self.prj)

        # when
        self.run_build('python setup.py build_project --build-package')

        # then
        self.assertTrue(python_package_built())
        self.assertTrue(is_test_run())
        self.assertFalse(package_leftovers_exist())


class BuildDagsCommandE2E(SetupTestCase):

    def test_should_execute_build_dags_command(self):

        # given
        create_dags_leftovers(self.prj.project_dir)
        clear_image_leftovers(self.prj)
        clear_package_leftovers(self.prj)

        # when
        self.run_build('python setup.py build_project --build-dags')

        # then
       # self.assertTrue(dags_built(TEST_PROJECT_PATH, 2))
        self.assertFalse(dags_leftovers_exist(TEST_PROJECT_PATH))
        self.assertTrue(dags_contain(TEST_PROJECT_PATH, repr(datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0) - timedelta(hours=24))))

        # when
        self.run_build(["python", "setup.py", "build_project", "--build-dags", "--start-time", "2020-01-02 00:00:00"])

        # then
        self.assertTrue(dags_contain(TEST_PROJECT_PATH, 'datetime.datetime(2020, 1, 1, 0, 0)'))

        # when
        self.run_build('python setup.py build_project --build-dags --workflow workflow1')

        # then
        self.assertTrue(self.single_dag_for_workflow_exists('workflow1'))

    def single_dag_for_workflow_exists(self, workflow_id):
        dags_created = dir_not_empty(TEST_PROJECT_PATH / '.dags')
        if not dags_created:
            return False
        dag_file_path = os.path.join(*next(walk_module_files(TEST_PROJECT_PATH / '.dags')))
        with open(dag_file_path, 'r') as f:
            return workflow_id in f.read()


class BuildImageCommandE2E(SetupTestCase):

    def test_should_execute_build_image_command(self):

        # given
        create_image_leftovers(self.prj.project_dir)
        clear_dags_leftovers(self.prj)
        clear_package_leftovers(self.prj)

        self.run_build('python setup.py build_project --build-package')

        # when
        self.run_build('python setup.py build_project --build-image')

        # then
        self.assertFalse(image_leftovers_exist())
        self.assertFalse(docker_image_built_in_registry(DOCKER_REPOSITORY, '0.1.0'))
        self.assertTrue(docker_image_as_file_built())
        self.assertTrue(deployment_config_copied())


class BuildImageTest(SetupTestCase):

    @mock.patch('bigflow.commons.build_docker_image_tag')
    @mock.patch('bigflow.build.operate._build_docker_image')
    @mock.patch('bigflow.build.operate._export_docker_image_to_file')
    @mock.patch('bigflow.commons.remove_docker_image_from_local_registry')
    def test_should_remove_image_from_local_registry_when_export_to_image_failed(self,
                                                                                 remove_docker_image_from_local_registry,
                                                                                 export_docker_image_to_file,
                                                                                 build_docker_image: mock.Mock,
                                                                                 build_docker_image_tag):
        # given
        image_dir_path = Path('image_dir')
        rmdir(image_dir_path)
        export_docker_image_to_file.side_effect = Exception('export failed')
        build_docker_image_tag.return_value = 'tag_value'

        # when:
        with self.assertRaises(Exception):
            build_image(self.prj)

        # then
        build_docker_image.assert_called_with(self.prj.project_dir, 'tag_value')
        remove_docker_image_from_local_registry.assert_called_with('tag_value')



"""This module is not part of the public API. No backwards compatibility guaranteed."""

import typing as tp
import abc
import logging
from bigflow.commons import run_process

__all__ = [
    'try_create',
    'CloudComposer'
]

logger = logging.getLogger(__name__)


class _InfrastructureManager(metaclass=abc.ABCMeta):
    name: str

    BIGFLOW_INFRA_PREFIX = 'bigflow'

    @abc.abstractmethod
    def create(self) -> None:
        pass

    @abc.abstractmethod
    def destroy(self) -> None:
        pass

    def secure_destroy(self) -> None:
        logger.info(f'Destroying infrastructure: {self.name}')
        try:
            self.destroy()
        except Exception as e:
            logger.exception(f"Couldn't destroy infrastructure: {self.name}", exc_info=e)


class _CloudRouter(_InfrastructureManager):
    def __init__(
            self,
            gcp_project_id: str,
            bigflow_project_name: str,
            environment_name: str,
            region: str):
        self.gcp_project_id = gcp_project_id
        self.router_name = f'{self.BIGFLOW_INFRA_PREFIX}-router-{bigflow_project_name}-{environment_name}-{region}'
        self.region = region
        self.name = self.router_name

    def create(self) -> None:
        run_process([
            'gcloud', 'compute',
            '--project', self.gcp_project_id,
            'routers', 'create', self.router_name,
            '--network', 'default',
            '--region', self.region
        ])

    def destroy(self) -> None:
        run_process([
            'gcloud', 'compute',
            '--project', self.gcp_project_id,
            'routers', 'delete', self.router_name,
            '--region', self.region,
            '--quiet'
        ])


class _CloudNat(_InfrastructureManager):
    def __init__(
            self,
            gcp_project_id: str,
            bigflow_project_name: str,
            environment_name: str,
            region: str):
        self.gcp_project_id = gcp_project_id
        self.nat_name = f'{self.BIGFLOW_INFRA_PREFIX}-nat-{bigflow_project_name}-{environment_name}-{region}'
        self.name = self.nat_name
        self.router = _CloudRouter(gcp_project_id, bigflow_project_name, environment_name, region)

    def create(self) -> None:
        self.router.create()
        run_process([
            'gcloud', 'compute',
            '--project', self.gcp_project_id,
            'routers', 'nats', 'create', self.nat_name,
            f'--router={self.router.name}',
            '--auto-allocate-nat-external-ips',
            '--nat-all-subnet-ip-ranges',
            '--enable-logging'
        ])

    def destroy(self) -> None:
        run_process([
            'gcloud', 'compute',
            '--project', self.gcp_project_id,
            'routers', 'nats', 'delete', self.nat_name,
            '--router', self.router.name,
            '--quiet'
        ])
        self.router.secure_destroy()


class CloudComposer(_InfrastructureManager):
    def __init__(
            self,
            gcp_project_id: str,
            bigflow_project_name: str,
            environment_name: str = 'dev',
            region: str = 'europe-west1',
            zone: str = 'europe-west1-d'):
        self.composer_name = f'{self.BIGFLOW_INFRA_PREFIX}-composer-{bigflow_project_name}-{environment_name}-{region}'
        self.gcp_project_id = gcp_project_id
        self.region = region
        self.zone = zone
        self.environment_name = environment_name
        self.cloud_nat = _CloudNat(gcp_project_id, bigflow_project_name, environment_name, region)
        self.name = self.composer_name

    def create(self) -> None:
        self.cloud_nat.create()
        run_process(_composer_create_command(
            self.composer_name,
            self.gcp_project_id,
            self.region,
            self.zone,
            self.environment_name))

    def destroy(self) -> None:
        run_process([
            'gcloud', 'composer',
            'environments', 'delete', self.composer_name,
            '--location', self.region,
            '--project', self.gcp_project_id,
            '--quiet'
        ])
        self.cloud_nat.secure_destroy()


def _composer_create_command(
        composer_name: str,
        gcp_project_id: str,
        region: str,
        zone: str,
        environment_name: str) -> tp.List[str]:
    return [
        'gcloud', 'composer',
        '--project', gcp_project_id,
        'environments', 'create', composer_name,
        f'--location={region}',
        f'--zone={zone}',
        f'--env-variables=env={environment_name}',
        f'--machine-type=n1-standard-2',
        f'--node-count=3',
        f'--python-version=3',
        f'--enable-ip-alias',
        f'--network=default',
        f'--subnetwork=default',
        f'--enable-private-environment',
        f'--enable-private-endpoint',
    ]


def try_create(infrastructure_element: _InfrastructureManager) -> None:
    try:
        infrastructure_element.create()
    except Exception:
        logger.exception(f"Can't create {infrastructure_element.name}. Trying to destroy leftovers.")
        infrastructure_element.secure_destroy()


if __name__ == '__main__':
    import argparse

    infra_cli_parser = argparse.ArgumentParser(description='Infrastructure management CLI')
    infra_cli_parser.add_argument('--gcp_project_id',
                                  type=str,
                                  help='The Google Cloud Platform project ID '
                                       'that you want to use as a place for the new Composer')
    infra_cli_parser.add_argument('--bigflow_project_name',
                                  type=str,
                                  help='The BigFlow project name for which the new Composer will be used')
    infra_cli_parser.add_argument('--region',
                                  type=str,
                                  default='europe-west1',
                                  help='The Google Cloud Platform region that you want to use for your Composer.')
    infra_cli_parser.add_argument('--zone',
                                  type=str,
                                  default='europe-west1-d',
                                  help='The Google Cloud Platform zone that you want to use for your Composer.')
    infra_cli_parser.add_argument('--environment',
                                  type=str,
                                  default='dev',
                                  help='The name for your environment, like dev/test/prod/etc.')
    args = infra_cli_parser.parse_args()

    try_create(CloudComposer(
        args.gcp_project_id,
        args.bigflow_project_name,
        args.environment,
        args.region,
        args.zone
    ))



### EXPERIMENTAL

import typing
import logging
import abc
import os

import pandas as pd

import apache_beam as beam


from . import io


logger = logging.getLogger(__name__)


def load_catboost_classifier(path: str, **kwargs):
    """Loads CatBoost classifier from single binary file (unpacked)"""

    import catboost
    logger.info("Load CatBoost model from %s, params %s", path, kwargs)

    m = catboost.CatBoostClassifier(**kwargs)
    with io.download_file_to_localfs(path) as fn:
        m.load_model(fn)

    return m


class BaseModel(abc.ABC):

    def __init__(self, model_path, model_opts=None, predict_opts=None):
        self.model_path = model_path
        self.model_opts = model_opts or {}
        self.predict_opts = predict_opts or {}
        self._model = None

    @abc.abstractmethod
    def load_model(self, path, **kwargs):
        raise NotImplementedError

    def ensure_model(self) -> typing.Any:
        if not self._model:
            self._model = self.load_model(self.model_path, **self.model_opts)
        return self._model

    def predict(self, X: pd.DataFrame):
        return pd.DataFrame(self.ensure_model().predict(X, **self.predict_opts))

    def __getstate__(self):
        # Don't tranfer model over network (pickling might not work for big models)
        d = self.__dict__.copy()
        d['_model'] = None
        return d


class CatBoostClassifierModel(BaseModel):

    def load_model(self, path, **kwargs):
        return load_catboost_classifier(path, **kwargs)

    def predict(self, X: pd.DataFrame):
        m = self.ensure_model()
        X = X.reset_index()
        Xs = X[m.feature_names_]
        return super().predict(Xs)


class MlOpsModelPredictor(BaseModel):

    def __init__(self, model_type, **kwargs):
        super().__init__(**kwargs)
        self.model_type = model_type
        assert not self.predict_opts

    def load_model(self, path, **kwargs):
        logger.info("Load mlops model %s from %s, opts %s", self.model_type, path, kwargs)
        m = self.model_type(**kwargs)
        with io.download_and_unpack_to_localfs(path) as d:
            logger.info("Load model from dir %s", d)
            m.load(d)
        return m

    def predict(self, X):
        request = {
            'instances': [
                {'values': list(x)}
                for x in X.values
            ],
        }
        response = self.ensure_model().predict(request)
        y = response['predictions']
        return pd.DataFrame(y)


def apply_model(df: pd.DataFrame, model: 'BaseModel', key_column: typing.Union[str, typing.List[str]]):
    logger.info("Process chunk, size %d", len(df))
    df = df.reset_index()
    i = df[key_column]
    y = model.predict(df)
    return pd.concat([i, y], axis=1)


@beam.ptransform_fn
def ApplyModel(
    pcoll: beam.PCollection[pd.DataFrame],
    model,
    key_column,
):
    return (pcoll
        | "Apply model"   >> beam.Map(apply_model, model=model, key_column=key_column)
    )


import typing
import logging
import uuid
import inspect

from apache_beam import Pipeline
from apache_beam.options.pipeline_options import PipelineOptions
from apache_beam.runners.runner import PipelineResult, PipelineState

from typing import Dict, List, Union, Optional, Tuple

try:
    from typing import TypedDict, Literal
except ImportError:
    from typing_extensions import TypedDict, Literal

from bigflow.commons import public, build_docker_image_tag
from bigflow.workflow import Job, JobContext

import bigflow.build.reflect


logger = logging.getLogger(__file__)

# PipelineOptions may dump unknown arguments to log which may cause passwords/secrets leakage.
logging.getLogger("apache_beam.options.pipeline_options").setLevel(logging.ERROR)


class PipelineOptionsDict(TypedDict, total=False):
    """Standard dataflow pipeline options.

    See sources for full list of supported options.
    https://github.com/apache/beam/blob/master/sdks/python/apache_beam/options/pipeline_options.py
    """

    runner: Literal['DirectRunner', 'DataflowRunner']
    experiments: List[str]
    streaming: bool

    job_name: str
    labels: List[str]

    temp_location: str
    staging_location: str

    project: str
    region: str
    zone: str
    worker_region: str
    worker_zone: str

    autoscaling_algorithm: Literal['THROUGHPUT_BASED', 'NONE']
    machine_type: str
    num_workers: int
    max_workers: int

    network: str
    subnetwork: str
    use_public_ips: bool
    service_account_email: str
    worker_disk_type: str
    disk_size_gb: int
    flexrs_goal: str


@public()
class BeamJob(Job):

    pipeline_level_execution_timeout_shift = 120  # 2 minutes

    def __init__(
            self,
            id: str = None,

            entry_point: typing.Callable = None,
            entry_point_args: Optional[Tuple] = None,
            entry_point_kwargs: Optional[Dict] = None,
            entry_point_arguments: typing.Optional[dict] = None,

            pipeline_options: Union[PipelineOptionsDict, PipelineOptions, None] = None,
            pipeline_options_no_bigflow_defaults: bool = False,

            wait_until_finish: bool = True,
            test_pipeline: Pipeline = None,
            execution_timeout_sec: int = None,
            use_docker_image: typing.Union[str, bool] = False,
            project_name=None,
            **kwargs,
    ):
        if bool(test_pipeline) == bool(pipeline_options is not None):
            raise ValueError("One of the pipeline and pipeline_options must be provided.")

        if not wait_until_finish and execution_timeout_sec:
            raise ValueError("If wait_until_finish_set to False execution_timeout can not be used.")

        super().__init__(
            id=id,
            execution_timeout_sec=execution_timeout_sec,
            **kwargs,
        )

        if isinstance(pipeline_options, PipelineOptions):
            logger.info("Convert PipelineOptions to dict")
            orig = pipeline_options
            pipeline_options = pipeline_options.get_all_options(
                drop_default=True
            )
            new = PipelineOptions(flags=[], **pipeline_options)
            assert orig.get_all_options() == new.get_all_options(), "During convertsion PipelineOptions<>dict some parameters was gone"

        assert not (pipeline_options_no_bigflow_defaults and use_docker_image), \
            "Option `use_docker_image` implies adding custom option, however `pipeline_options_no_bigflow_defaults` is specified"

        self.pipeline_options = dict(pipeline_options or {})

        self.wait_until_finish = wait_until_finish
        self.test_pipeline = test_pipeline
        self.use_docker_image = use_docker_image
        self.entry_point = entry_point
        self.pipeline_options_no_bigflow_defaults = pipeline_options_no_bigflow_defaults

        self._project_path = bigflow.build.reflect.locate_project_path(project_name)

        if (entry_point_arguments is None
            and entry_point_args is None
            and entry_point_kwargs is None
        ):
            try:
                inspect.signature(entry_point).bind(None, None, None)
            except TypeError:
                pass
            else:
                logger.warning("Passing empty {} as `entry_point_arguments` %s - you can drop this unused argumnent", entry_point)
                entry_point_arguments = {}

        if entry_point_arguments is None:
            # threading-like - tuple/dict for args/kwargs
            self.entry_point_args = entry_point_args or ()
            self.entry_point_kwargs = entry_point_kwargs or {}
            self.entry_point_arguments = None

        else:
            # old style - single positional argument with type 'dict'
            logger.warning("Please use `entry_point_kwargs` instead of `entry_point_arguments`")
            assert entry_point_args is None, "Mixing `entry_point_args` and `entry_point_arguments` is not allowed"
            assert entry_point_kwargs is None, "Mixing `entry_point_kwargs` and `entry_point_arguments` is not allowed"
            self.entry_point_args = (entry_point_arguments,)
            self.entry_point_kwargs = {}
            self.entry_point_arguments = entry_point_arguments

    def execute(self, context: JobContext):
        pipeline = self.test_pipeline or self.new_pipeline(context)

        logger.info("init beam pipeline...")
        self.init_pipeline(context, pipeline)

        logger.info("run beam pipeline...")
        result = self.run_pipeline(context, pipeline)

        logger.info("wait pipeline result...")
        self.wait_pipeline_result(result)

    def wait_pipeline_result(self, result: PipelineResult):
        if self.wait_until_finish and self.execution_timeout_sec:
            timeout_in_milliseconds = 1000 * (self.execution_timeout_sec - self.pipeline_level_execution_timeout_shift)
            result.wait_until_finish(timeout_in_milliseconds)
            if not PipelineState.is_terminal(result.state):
                result.cancel()
                raise RuntimeError(f'Job {self.id} timed out ({self.execution_timeout_sec})')

    def new_pipeline(self, context: JobContext) -> Pipeline:
        logger.debug("Create new pipline for context %s", context)
        popts = self.create_pipeline_options(context)
        return Pipeline(options=popts)

    def init_pipeline(self, context: JobContext, pipeline: Pipeline):
        if not self.entry_point:
            raise RuntimeError("You need override method 'init_pipeline' or provide 'entry_point' argument")
        self.entry_point(pipeline, context, *self.entry_point_args, **self.entry_point_kwargs)

    def run_pipeline(self, context: JobContext, pipeline: Pipeline):
        logger.info("Run pipeline, context %s", context)
        return pipeline.run()

    def create_pipeline_options(self, context: JobContext) -> PipelineOptions:
        options = dict(self.pipeline_options)
        if not self.pipeline_options_no_bigflow_defaults:
            self.set_default_pipeline_options(context, options)
        else:
            logger.debug("Don't any any defaults to pipeline options")

        logger.debug("Pipeline options from dict %s", options)
        return PipelineOptions(flags=[], **options)

    def set_default_pipeline_options(self, context: JobContext, options: PipelineOptionsDict):

        logger.info("Add defaults to pipeline options")
        options.setdefault('runner', 'DataflowRunner')

        if 'job_name' not in options:
            slug = self.id.replace("_", "-")
            job_name = f"{slug}-{uuid.uuid4().hex}"
            logger.info("Use default job name %r", job_name)
            options['job_name'] = job_name
        else:
            logger.info("Keep provided job name %r", options['job_name'])

        if options['runner'].upper() != 'DATAFLOWRUNNER':
            logger.debug("Don't add default setup.py for runner %r", options['runner'])
        elif 'setup_file' not in options:
            setuppy = str(bigflow.build.reflect.materialize_setuppy(self._project_path))
            logging.debug("Add setup.py %s to pipeline options", setuppy)
            options['setup_file'] = setuppy
        else:
            logging.debug("Pipeline options already contains 'setup_file': %s", options['setup_file'])

        workflow_id = context.workflow_id
        if workflow_id:
            options['labels'] = (options.get('labels') or []) + [f'workflow_id={workflow_id}']
        else:
            logger.warning("A workflow_id is not found in the context - skip logging initialization.")

        if self.use_docker_image:
            if isinstance(self.use_docker_image, str):
                logger.debug("Use explicitly provided docker image")
                imgid = self.use_docker_image
            else:
                logger.debug("Infer docker image name for current project")
                pspec = bigflow.build.reflect.get_project_spec(self._project_path)
                imgid = build_docker_image_tag(pspec.docker_repository, pspec.version)
            logger.info("Use docker image %s for beam workers", imgid)

            options['worker_harness_container_image'] = imgid
            experiments = list(options.get('experiments', []))
            if 'use_runner_v2' not in experiments:
                logger.info("Enable beam experiment 'use_runner_v2'")
                experiments.append('use_runner_v2')
            options['experiments'] = experiments



from bigflow.commons import public

from . import interactive, dataset_configuration, job
from .interactive import INLINE_COMPONENT_DATASET_ALIAS

__all__ = [
    'component',
    'add_label',
    'sensor',
    'INLINE_COMPONENT_DATASET_ALIAS',
    'DatasetConfig',
    'Job'
]


@public(class_alias=True)
class DatasetConfig(dataset_configuration.DatasetConfig): ...


@public(class_alias=True)
class Job(job.Job): ...


@public(alias_for=interactive.add_label)
def add_label(table_name, labels, ds=None): ...


@public(alias_for=interactive.sensor)
def sensor(table_alias, where_clause, ds=None): ...


@public(alias_for=interactive.interactive_component)
def component(**dependencies): ...



#!/usr/bin/evn python

# Crawl Dataflow documentation, generates requirements files for various beam/python combinations.
# Requirements: `requests`, `beautifulsoup4`

import requests
import bs4
import re
import os


BEAM_VERSIONS = {
    "2.24", "2.25", "2.26", "2.27", "2.28",
}


def grab_sdk_worker_dependencies():

    page = requests.get("https://cloud.google.com/dataflow/docs/concepts/sdk-worker-dependencies")
    page.raise_for_status()

    bs = bs4.BeautifulSoup(page.text, features='html.parser')

    for section in bs.find_all('section'):
        if not re.match(r"py\d+", section.get('id', "")):
            continue

        beam_ver = section.h4['id'].split("-")[-1]
        beam_ver = ".".join(beam_ver.split(".")[:2])  # only major_minor

        for sub_section in section.find_all('section'):
            python_version = sub_section.h3['id'].split("-")[-1]
            python_version = ".".join(python_version.split(".")[:2])  # only major_minor

            packages = {}
            for tr in sub_section.find_all('tr'):
                if not tr.td and tr.th:
                    # header
                    continue
                package_name = tr.td.text
                version = tr.td.next_sibling.text
                packages[package_name] = version

            yield {
                'python_version': python_version,
                'beam_version': beam_ver,
                'packages': packages,
            }


def main():

    all_worker_dependencies = list(grab_sdk_worker_dependencies())
    print("python versions", sorted(set(worker_dependencies['python_version'] for worker_dependencies in all_worker_dependencies)))
    print("beam versions", sorted(set(worker_dependencies['beam_version'] for worker_dependencies in all_worker_dependencies)))

    for worker_dependencies in all_worker_dependencies:
        beam_version = worker_dependencies['beam_version']
        if beam_version not in BEAM_VERSIONS:
            print("skip beam version", beam_version)
            continue

        python_version = worker_dependencies['python_version']
        if python_version.startswith("2"):
            continue

        pkgs = worker_dependencies['packages']

        out_file_name = f"beam{beam_version}_py{python_version}.txt"
        work_dir = os.path.join(os.path.dirname(__file__))

        with open(os.path.join(work_dir, out_file_name), 'w') as f:
            print(f.name)

            f.write(f"# list of preinstalled dependencies on Dataflow workers\n")
            f.write(f"# this file was generated by `grab_beam_worker_deps.py`\n")
            f.write(f"# beam sdk version: {beam_version}\n")
            f.write(f"# python version: {python_version}\n")
            f.write("\n")

            for worker_dependencies, v in sorted(pkgs.items()):
                f.write(f"{worker_dependencies}=={v}\n")


if __name__ == '__main__':
    main()


"""Mixins to isolate `unittest.TestCase` into separate processes

May prevent from doing side effects with Python intepreter
(loaded modules, global variables, caches etc).
"""

# `*IsolateMixin`s overrides original `unittest.TestCase.run` method.
# It spawns a new subprocess via `multiprocessing`.
# With 'spawn' strategy the whole unittest is pickled and passend into subprocess.
# With 'fork' strategy pickling is not involved.
#
# Then parent creates proxy to instance of `unittest.TestResult` and pass proxy to
# child process.  Such synchronously sends information about all methods calls
# to parent via duplex pipe.  Parent recieves information and mutate
# original 'TestResult' appropriately.
#
# You can find sequence diagramm for this in "isolate_sequence.puml" file.


import unittest
import sys
import time
import contextlib
import multiprocessing
import threading


from bigflow.commons import public


class _ObjectTag:
    """Stub for non-pickable objects, which needed to be transferred between processes"""

    def __init__(self, tag):
        self.tag = tag

    @classmethod
    def forfeit(cls, obj, obj2tag):
        return cls(obj2tag[obj]) if obj in obj2tag else obj

    @classmethod
    def recover(cls, obj, tag2obj):
        return tag2obj[obj.tag] if isinstance(obj, cls) else obj


class _TestResultProxy:
    """Dump proxy for `TestResult`. Stream all public method calls into channel."""

    def __init__(self, ch, testcase_proxy, tblib):
        self._ch = ch
        self._tc_proxy = testcase_proxy
        self._tblib = tblib
        self._lock = threading.Lock()

    # proxy all public method calls
    def __getattr__(self, name: str):
        assert not name.startswith("_")

        def method(*args):
            with self._lock:
                # replace ref to proxy testcase by sepulka
                smap = {self._tc_proxy: 'testcase'}
                args = [_ObjectTag.forfeit(a, smap) for a in args]

                if isinstance(args[-1], tuple) and not self._tblib:
                    # last arg is 'exc_info', but `tblib` is not installed
                    args[-1] = (None, None, None)

                self._ch.send((name, tuple(args)))
                return self._ch.recv()

        return method


class _IsolatedProcessMixin(unittest.TestCase):

    timeout: float
    isolate: bool

    _was_executed: bool = False
    _isolated_test_spawn_method: str

    def _enable_tblib(self):
        try:
            import tblib.pickling_support
        except ImportError:
            print("Failed to import `tblib`, tracebask/stacktracs might be stripped")
            return False
        else:
            tblib.pickling_support.install()
            return True

    def _stop_process(self, p):
        if p.is_alive():
            p.kill()
            p.join(timeout=30)
        p.close()

    def _fail_into_result(self, result: unittest.TestResult, msg):
        """Include error with the traceback into `TestResult`"""
        try:
            raise RuntimeError(msg)
        except RuntimeError:
            result.addFailure(self, sys.exc_info())

    def run0(self, ch, tblib_enabled):
        # executed from child process
        tblib_enabled = tblib_enabled and self._enable_tblib()
        with ch:
            super().run(_TestResultProxy(ch, self, tblib_enabled))
            ch.send(('__stop__', None))

    def _recv_testresult_proxy(self, result, process, ch, run_until):
        """Recieve and applly method calls on real TestResult object"""

        poll_delay = 0.001
        while process.is_alive() and time.time() < run_until:
            if not ch.poll(poll_delay):
                poll_delay = min(1, poll_delay * 1.5)
                continue

            poll_delay = 0.001
            method, args = ch.recv()
            if method == '__stop__':
                return  # ok

            # recover ref to current testcase
            smap = {'testcase': self}
            args = tuple(_ObjectTag.recover(a, smap) for a in args)

            resp = getattr(result, method)(*args)
            ch.send(resp)

        if process.is_alive():
            self._fail_into_result(result, "Timeout: isolated test subprocess is stil running")
        else:
            self._fail_into_result(result, "Isotalted process was unexpectedly terminated")

    def run(self, result=None):
        cls = type(self)

        if not self.isolate:
            assert not cls._was_executed, "Test was already executed without isolation, either enable isolation or run single test at once"
            try:
                self.setUpParent()
                return super().run(result)
            finally:
                self.tearDownParent()

        cls._was_executed = True
        tblib_enabled = self._enable_tblib()

        if result is None:
            result = self.defaultTestResult()
        mp_context = multiprocessing.get_context(
            self._isolated_test_spawn_method)

        self.setUpParent()
        with contextlib.ExitStack() as cstack:
            cstack.callback(self.tearDownParent)

            server_ch, client_ch = mp_context.Pipe()
            cstack.callback(server_ch.close)
            cstack.callback(client_ch.close)

            process = mp_context.Process(
                name=f"isolated-test--{self.id}",
                target=self.run0, args=[client_ch, tblib_enabled],
            )
            process.start()
            cstack.callback(self._stop_process, process)

            run_until = time.time() + self.timeout
            self._recv_testresult_proxy(result, process, server_ch, run_until)
            process.join(max(1, run_until - time.time()))

    @property
    def isolate(self):
        # isolate only when running without debugger
        return sys.gettrace() == None

    def setUpParent(self):
        "Hook method for setting up the test fixture before exercising it. Runs in original python process."
        pass

    def tearDownParent(self):
        "Hook method for deconstructing the test fixture after testing it. Runs in original python process."
        pass


@public()
class ForkIsolateMixin(_IsolatedProcessMixin):
    """Mixin class for subclasses of `unittest.TestCase`

    Run all test methods in separate forked processes.
    All changes in variables/modules/memory made during test execution does not affect original Python instance.
    Note, that IO operations are not isolated, so external resources still need to be cleaned manually.

    Methods `setUpClass` and `tearDownClass` are executed in scope of main process.

    Methods `setUp` and `tearDown` are executed inside spawned process.
    Additional methods `setUpParent` and `tearDownParent` provided to run
    per-test fixtures in scope of parent process.
    """

    timeout: float = 3600
    isolate: bool

    _isolated_test_spawn_method = 'fork'


@public()
class SpawnIsolateMixin(_IsolatedProcessMixin):
    """Mixin class for subclasses of `unittest.TestCase`

    Run all test methods in fresh Python processes.
    Each process is inializated 'from scratch', which result into recreating/reloading all modules/classes.
    Note, that IO operations are not isolated, so external resources still need to be cleaned manually.

    Methods `setUpClass` and `tearDownClass` are executed in scope of main process,
    however it is not recommended to use them with this mixin.

    Methods `setUp` and `tearDown` are executed inside spawned process.
    Additional methods `setUpParent` and `tearDownParent` provided to run
    per-test fixtures in scope of parent process.

    Test instance and test class must be pickleable.
    """

    timeout: float = 3600
    isolate: bool

    _isolated_test_spawn_method = 'spawn'

    def __init__(self, *args, **kwargs):
        self._class_dict_orig = dict(self.__class__.__dict__)
        super().__init__(*args, **kwargs)

    # Pickle not only test instance state, but also state of test class
    # Only fields changed after test `__init__` are tranferred into subprocess.
    def __getstate__(self):
        selfdict = dict(self.__dict__)
        clsdict_orig = selfdict.pop('_class_dict_orig')
        clsdict = {
            k: v
            for k, v in self.__class__.__dict__.items()
            if k not in clsdict_orig or clsdict_orig[k] is not v
        }
        return selfdict, clsdict

    def __setstate__(self, state):

        selfdict, clsdict = state
        self.__dict__ = selfdict

        # dynamically overwrite test class with new state
        pcls = type(self.__class__.__name__, (self.__class__,), clsdict)
        self.__class__ = pcls



import typing
from pathlib import Path
from abc import ABC, abstractmethod
from .job import DEFAULT_RETRY_COUNT
from .job import DEFAULT_RETRY_PAUSE_SEC

from bigflow.commons import public

DEFAULT_RUNTIME = '1970-01-01'


@public()
class BigQueryOperation(ABC):

    @abstractmethod
    def to_job(self, id: str, retry_count: int = DEFAULT_RETRY_COUNT, retry_pause_sec: int = DEFAULT_RETRY_PAUSE_SEC):
        pass

    @abstractmethod
    def run(self, runtime=DEFAULT_RUNTIME):
        pass


@public()
class Dataset(ABC):

    @abstractmethod
    def write_truncate(self, table_name: str, sql: str, partitioned: bool = True) -> BigQueryOperation:
        pass

    @abstractmethod
    def write_append(self, table_name: str, sql: str, partitioned: bool = True) -> BigQueryOperation:
        pass

    @abstractmethod
    def write_tmp(self, table_name: str, sql: str) -> BigQueryOperation:
        pass

    @abstractmethod
    def collect(self, sql: str) -> BigQueryOperation:
        pass

    @abstractmethod
    def collect_list(self, sql: str, record_as_dict: bool = False) -> BigQueryOperation:
        pass

    @abstractmethod
    def create_table(self, create_query: str) -> BigQueryOperation:
        pass

    @abstractmethod
    def create_table_from_schema(
            self,
            table_name: str,
            schema: typing.Union[typing.List[dict], Path, None] = None,
            table=None):
        pass

    @abstractmethod
    def insert(
            self,
            table_name: str,
            records: typing.Union[typing.List[dict], Path]):
        pass

    @abstractmethod
    def delete_dataset(self):
        pass


import abc
import collections
import collections.abc
import logging
import os
import re
import typing
import sys

from typing import (
    Any,
    Callable,
    Generic,
    List,
    Optional,
    Tuple,
    Dict,
    Type,
)

import lazy_object_proxy  # type: ignore
from bigflow.commons import public

__all__ = [
    'dynamic',
    'expand',
    'fromenv',
    'Konfig',
    'resolve_konfig',
]


logger = logging.getLogger(__name__)
K = typing.TypeVar('K', bound='Konfig')
T_co = typing.TypeVar('T_co', covariant=True)


if sys.version_info >= (3, 9):
    from functools import cached_property
else:
    # python 3.8 has 'cached_property', but it is not subscriptable
    from backports.cached_property import cached_property as _cached_property
    class cached_property(Generic[T_co], _cached_property):
        def __get__(self, instance: Any, owner: Optional[Type]) -> T_co:   # type: ignore
            return super().__get__(instance, owner)


def current_env() -> Optional[str]:
    """Returns current env name (specified via 'bigflow --config' option)"""
    return os.environ.get('bf_env')


class KonfigMeta(abc.ABCMeta):

    @staticmethod
    def _copy_classvars_to_obj(klass, obj):
        for k, v in klass.__dict__.items():
            if k.startswith("_"):
                # ignore private attributes
                continue
            elif hasattr(v, '__get__'):
                # descriptor, pop out attribute from obj
                obj.__dict__.pop(k, None)
            else:
                # normal value, put into obj
                obj.__dict__[k] = v

    def __new__(metacls, clsname, bases, dct):

        if not any(isinstance(b, metacls) for b in bases):
            # no Konfig's in bases - create root type
            return type.__new__(metacls, clsname, bases, dct)

        if '__init__' in dct:
            raise ValueError("Konfig classes doesn't support custom '__init__'")

        def __init__(self, **kwargs):
            super(cls, self).__init__(**kwargs)
            metacls._copy_classvars_to_obj(cls, self)

        cls = type.__new__(metacls, clsname, bases, {'__init__': __init__, **dct})
        return cls

    def __call__(self, **kwargs):
        obj = super().__call__(**kwargs)
        obj.__dict__.update(**kwargs)
        obj.__post_init__(**kwargs)
        obj._frozen = True
        return obj


@public()
class Konfig(collections.abc.Mapping, metaclass=KonfigMeta):
    """Base class for configs.

    Automatically copies all class-level attributes into `self` in class definition order.
    Protects constructed object from mutation (see `_frozen` attribute).
    """

    __slots__ = ('__dict__', '_frozen', '_init_kwargs')

    def __init__(self, **kwargs):
        self._init_kwargs = kwargs

    def __post_init__(self, **kwargs):
        for k in dir(self):
            if not k.startswith("_"):
                # "materialize" all 'cached_property's
                getattr(self, k)

    def __setattr__(self, name, value):
        if getattr(self, '_frozen', False) and name != '_frozen':
            raise RuntimeError("Attempt to modify frozen Konfig")
        object.__setattr__(self, name, value)

    def __repr__(self):
        cls = type(self)
        kvpairs_list = ", ".join(f"{k}={v!r}" for k, v in self.items())
        return f"{cls.__module__}.{cls.__qualname__}({kvpairs_list})"

    @classmethod
    def _make(cls: Type[K], kvpairs: List[Tuple[str, Any]]) -> K:
        return cls(**dict(kvpairs))

    def __reduce__(self):
        return (
            self._make,
            (list(self.__dict__.items()),),
        )

    def replace(self: K, **kwargs) -> K:
        """Return a new instance of the konfig replacing specified fields"""
        cls = type(self)
        return cls(**{**self._init_kwargs, **kwargs})

    # adapt to `collections.abc.Mapping`

    def __getitem__(self, k):
        return getattr(self, k)

    def __iter__(self):
        return iter(vars(self).keys())

    def __len__(self):
        return len(vars(self))


class _LazyKonfig(lazy_object_proxy.Proxy):

    def __init__(self, **kwargs):
        super().__init__(lambda: resolve_konfig(**kwargs))

    def __repr__(self):
        return str(self)


@public()
def resolve_konfig(
    konfigs: Dict[str, Type[K]],
    name: Optional[str] = None,
    default: Optional[str] = None,
    extra: Optional[Dict[str, Any]] = None,
    lazy: bool = True,
) -> K:

    """Creates instance of config.

    Konfigs are passed as a dictionary {name -> konfig_class}.

    Konfig name may be provided explicitly with `name` argument or passed via `bigflow` cli tool.
    Default konfig config name may be passed as a fallback.
    """

    konfigs = dict(konfigs)
    extra = dict(extra or {})

    if lazy:
        return _LazyKonfig(konfigs=konfigs, name=name, default=default, extra=extra, lazy=False)

    logger.debug("Resolve konfig %s from %s", name or "*", konfigs)
    name = name or current_env() or default

    if not name:
        raise ValueError("Konfig name should not be empty")
    logger.debug("Konfig name is %s", name)

    try:
        konfig_cls = konfigs[name]
    except KeyError:
        raise ValueError(f"Unable to find konfig with name{name!r}, candidates {list(konfigs.keys())}")

    logger.info("Create instance of konfig %s", konfig_cls)
    return konfig_cls(**extra)


class secretstr(str):
    def __repr__(self):
        return f"<secretstr {'*' * len(self)}>"


@public()
def fromenv(
    key: str,
    default: Optional[str] = None,
    type: Type = secretstr,
) -> cached_property[str]:

    """Reads config value from os environment, prepends `bf_` to variable name"""

    def __get__(self):
        try:
            raw = os.environ[key]
        except KeyError:
            if default is None:
                raise ValueError(f"No environment variable {key}")
            else:
                return type(default)
        else:
            return type(raw)

    return dynamic(__get__)


@public()
def expand(value: str) -> cached_property[str]:
    """Expands placeholders ('{key_name}' is replaced with value of `konfig.key_name`)"""

    def __get__(self):
        return _resolve_placeholders(
            value,
            lambda k: str(getattr(self, k))
        )

    return dynamic(__get__)


@public()
def dynamic(get: Callable[[K], T_co]) -> cached_property[T_co]:

    def __get__(self: K) -> T_co:
        assert isinstance(self, Konfig), f"object {self} must be subclass of `Konfig` class"
        return get(self)

    return cached_property(get)


_placeholder_re = re.compile(r"""
    \{ (\w+) \}   # placeholder `{key}`
    | \}\}        # literal `}}`
    | \{\{        # literal `}}`
""", re.VERBOSE)


def _resolve_placeholders(
    value: str,
    resolve: Callable[[str], Any],
) -> str:

    def valueof(m: re.Match) -> str:
        s = m.group(0)
        if s == "{{":
            return "{"
        if s == "}}":
            return "}"
        else:
            k = m.group(1)
            return str(resolve(k))

    return _placeholder_re.sub(valueof, value)



import json
import base64
import functools
import datetime
import time
import io
import pathlib
import pickle
import random
import string
import textwrap
import time
import typing
import contextlib
import logging

from google.cloud import dataproc_v1, storage

import bigflow
import bigflow.configuration
import bigflow.commons
import bigflow.build.reflect
import bigflow.build.pip
from bigflow.commons import public

from bigflow.workflow import DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS
from bigflow._version import __version__ as bf_version

logger = logging.getLogger(__name__)

__all__ = [
    'PySparkJob',
]


@public()
class PySparkJob(bigflow.Job):

    def __init__(
        self,
        id,
        driver: typing.Callable,
        bucket_id: str,
        gcp_project_id: str,
        gcp_region: str,
        driver_arguments: typing.Optional[dict] = None,
        pip_packages: typing.Union[typing.Iterable[str], pathlib.Path] = (f'bigflow[dataproc]=={bf_version}', ),
        jar_file_uris: typing.Iterable[str] = ('gs://spark-lib/bigquery/spark-bigquery-latest_2.12.jar',),
        worker_num_instances: int = 2,
        worker_machine_type: str = 'n1-standard-1',
        env: typing.Optional[str] = None,
        project_name: typing.Optional[str] = None,
        execution_timeout_sec: int = DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS,
        internal_ip_only: bool = False
    ):
        self.id = id

        if driver_arguments:
            driver = functools.partial(driver, **driver_arguments)

        self.driver = driver
        self.driver_filename = f"{id}__driver.py"

        self.bucket_id = bucket_id
        self.gcp_project_id = gcp_project_id
        self.gcp_region = gcp_region

        self.env = env or bigflow.configuration.current_env() or 'none'

        self.jar_file_uris = jar_file_uris
        self.worker_machine_type = worker_machine_type
        self.worker_num_instances = worker_num_instances

        if isinstance(pip_packages, pathlib.Path):
            self.pip_packages = bigflow.build.pip.read_requirements(pip_packages)
        else:
            self.pip_packages = pip_packages

        self._project_pkg_path = project_name or bigflow.build.reflect.locate_project_path(project_name)
        self.execution_timeout_sec = execution_timeout_sec
        self.internal_ip_only = internal_ip_only

    def _generate_internal_jobid(self, context):
        job_random_suffix = ''.join(random.choices(string.ascii_lowercase + string.digits, k=10))
        job_timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H%M%S")  # FIXME: use 'runtime' ?
        jobid = f"{self.id.lower()}-{self.env or 'none'}-{job_timestamp}-{job_random_suffix}"
        logger.info("Internal job is %r", jobid)
        return jobid.replace("_", "-")

    def _prepare_driver_callable(self, context: bigflow.JobContext):
        # Remove 'workflow' from context but keeps 'workflow_id'.
        # In most cases Workflow is not pickleable as it keeps references to custom user jobs.
        cleaned_context = context._replace(workflow=None)
        return functools.partial(self.driver, cleaned_context)

    def _prepare_driver_script(self, context: bigflow.JobContext):
        return generate_driver_script(
            self._prepare_driver_callable(context),
            self._prepare_env_variables(context),
        )

    @contextlib.contextmanager
    def _with_temp_cluster(self, cluster_name):
        client_options = {'api_endpoint': f"{self.gcp_region}-dataproc.googleapis.com:443"}
        dataproc_cluster_client = dataproc_v1.ClusterControllerClient(client_options=client_options)
        try:
            logger.debug("Create temp cluster %r", cluster_name)
            cluster_name = _create_cluster(
                dataproc_cluster_client=dataproc_cluster_client,
                project_id=self.gcp_project_id,
                region=self.gcp_region,
                cluster_name=cluster_name,
                requirements=self.pip_packages,
                worker_machine_type=self.worker_machine_type,
                worker_num_instances=self.worker_num_instances,
                internal_ip_only=self.internal_ip_only
            )
            yield cluster_name
        finally:
            logger.debug("Delete temp cluster %r", cluster_name)
            _delete_cluster(dataproc_cluster_client, self.gcp_project_id, self.gcp_region, cluster_name)

    def _prepare_env_variables(self, context):
        res = {}
        if context.env:
            res['bf_env'] = context.env
        if context.workflow and context.workflow.log_config:
            res['bf_log_config'] = json.dumps(context.workflow.log_config)
        if context.workflow_id:
            res['bf_workflow_id'] = context.workflow_id
        return res

    def _prepare_pyspark_properties(self, context):
        env = self._prepare_env_variables(context)
        ps = {
            'spark.app.name': self.id,
            **{f"spark.executorEnv.{k}": v for k, v in env.items()}
        }
        logger.debug("pyspark properties are %r", ps)
        return ps

    def execute(self, context: bigflow.JobContext):
        logger.info("Run job %r", self.id)

        job_internal_id = self._generate_internal_jobid(context)

        client_options = {'api_endpoint': f"{self.gcp_region}-dataproc.googleapis.com:443"}
        storage_client = storage.Client(project=self.gcp_project_id)
        dataproc_job_client = dataproc_v1.JobControllerClient(client_options=client_options)

        driver_script = self._prepare_driver_script(context)

        logger.info("Prapare and upload python package...")
        bucket = storage_client.get_bucket(self.bucket_id)

        egg_local_path = str(bigflow.build.reflect.build_egg(self._project_pkg_path))
        egg_path = _upload_egg(egg_local_path, bucket, job_internal_id)

        driver_path = f"{job_internal_id}/{self.driver_filename}"
        _upload_driver_script(driver_script, bucket, driver_path)

        with self._with_temp_cluster(job_internal_id) as cluster_name:
            job = _submit_single_pyspark_job(
                dataproc_job_client=dataproc_job_client,
                project_id=self.gcp_project_id,
                region=self.gcp_region,
                cluster_name=cluster_name,
                bucket_id=self.bucket_id,
                jar_file_uris=self.jar_file_uris,
                driver_path=driver_path,
                egg_path=egg_path,
                properties=self._prepare_pyspark_properties(context),
            )
            try:
                _wait_for_job_to_finish(dataproc_job_client, self.gcp_project_id, self.gcp_region, job)
            finally:
                _print_job_output_log(storage_client, dataproc_job_client, self.gcp_project_id, self.gcp_region, job)

        logger.info("Job %r was finished", self.id)


def generate_driver_script(callable, env):
    pickled = pickle.dumps(callable)
    env = dict(env or {})
    return textwrap.dedent(f"""
        # Generated by `bigflow.dataproc.generate_and_upload_driver`
        import os, base64, pickle

        # overwrite env variables
        os.environ.update({env!r})

        # triggers configuration of logging
        try:
            from bigflow.log import maybe_init_logging_from_env
        except ImportError:
            pass
        else:
            maybe_init_logging_from_env()

        # unpickle'n'call {callable!r}
        data = {base64.b85encode(pickled)!r}

        pickle.loads(base64.b85decode(data))()
    """)


def _upload_driver_script(script, bucket, driver_path):
    logger.debug("Upload generate driver to %s", driver_path)
    driver_blob = bucket.blob(driver_path)
    driver_blob.upload_from_string(content_type='application/octet-stream', data=script)


def _upload_egg(egg_local_path, bucket, job_internal_id):
    logger.debug("Upload egg to %s", egg_local_path)
    egg_path = '{}/{}'.format(job_internal_id, pathlib.Path(egg_local_path).name)
    egg_blob = bucket.blob(egg_path)
    egg_blob.upload_from_filename(filename=egg_local_path, content_type='application/octet-stream')
    return egg_path


def _wait_for_job_to_finish(dataproc_cluster_client, project_id, region, job_id):
    logger.info("Waiting for job to finish...")
    while True:
        job = dataproc_cluster_client.get_job(project_id=project_id, region=region, job_id=job_id)
        state = job.status.state
        if state == dataproc_v1.JobStatus.State.DONE:
            logger.info("Job is DONE")
            return job
        elif state == dataproc_v1.JobStatus.State.ERROR:
            logger.error("Job was failed with ERROR")
            raise Exception(job.status.details)
        elif state == dataproc_v1.JobStatus.State.CANCELLED:
            logger.warn("Job was CANCELLED")
            raise Exception("Job was CANCELLED")
        time.sleep(5)


def _print_job_output_log(storage_client, dataproc_cluster_client, project_id, region, job_id):

    job = dataproc_cluster_client.get_job(project_id=project_id, region=region, job_id=job_id)
    log_buffer = io.BytesIO()
    # FIXME(anjensan): Download all logs
    storage_client.download_blob_to_file(job.driver_output_resource_uri + ".000000000", log_buffer)
    logger.info("JOB OUTPUT:\n=====\n%s\n=====", log_buffer.getvalue().decode())


def _create_cluster(
        dataproc_cluster_client,
        project_id,
        region,
        cluster_name,
        requirements,
        worker_num_instances,
        worker_machine_type,
        internal_ip_only: bool = False):
    packages = " ".join(filter(None, requirements))
    cluster_data = {
        # "project_id": project_id,
        "cluster_name": cluster_name,
        "config": {
            "master_config": {
                "num_instances": 1,
                "machine_type_uri": "n1-standard-4",
                "disk_config": {
                    "boot_disk_size_gb": 100
                }},
            "worker_config": {
                "num_instances": worker_num_instances,
                "machine_type_uri": worker_machine_type,
                "disk_config": {
                    "boot_disk_size_gb": 100
                }},
            "software_config": {"image_version": "1.5"},
            "initialization_actions": [
                {"executable_file": "gs://goog-dataproc-initialization-actions-{}/python/pip-install.sh".format(region)}
            ],
            "gce_cluster_config": {
                "metadata": [('PIP_PACKAGES', packages)],
                "internal_ip_only": internal_ip_only
            }
        },
    }
    logger.debug("Cluster spec %r", cluster_data)

    start_at = time.time()
    logger.info("Create cluster %r", cluster_name)
    cluster_future = dataproc_cluster_client.create_cluster(
        project_id=project_id,
        region=region,
        cluster=cluster_data,
    )

    logger.info("Waiting for cluster creation...")
    cluster_future.result()

    passed = time.time() - start_at
    logger.info("Cluster created in %s seconds." % passed)

    return cluster_name


def _submit_single_pyspark_job(
        dataproc_job_client,
        project_id,
        region,
        cluster_name,
        bucket_id,
        driver_path,
        egg_path,
        jar_file_uris,
        properties,
    ):
    jobspec = {
        "placement": {
            "cluster_name": cluster_name,
        },
        "pyspark_job": {
            'main_python_file_uri': f"gs://{bucket_id}/{driver_path}",
            'jar_file_uris': jar_file_uris,
            'python_file_uris': [f"gs://{bucket_id}/{egg_path}"],
            'properties': properties,
        },
    }
    logger.debug("Pyspark job %r", jobspec)
    result = dataproc_job_client.submit_job(project_id=project_id, region=region, job=jobspec)

    job_id = result.reference.job_id
    print("Job {} submitted.".format(job_id))
    print(f"https://console.cloud.google.com/dataproc/jobs/{job_id}?project={project_id}&region={region}")

    return job_id


def _delete_cluster(dataproc_cluster_client, project_id, region, cluster):
    print("Deleting cluster...")
    dataproc_cluster_client.delete_cluster(
        project_id=project_id,
        region=region,
        cluster_name=cluster,
    )
    print("Cluster was deleted.")



import logging
import sys
import re
import toml

from typing import Optional
from pathlib import Path

import bigflow
import bigflow.build.dev
import bigflow.build.pip
import bigflow.build.reflect


logger = logging.getLogger(__name__)


def _yes_or_no() -> bool:
    yes = {'yes', 'y', 'ye', ''}
    corr = yes | {'no', 'n'}
    while True:
        choice = input("Please respond with y/yes/n/no [Y]").lower()
        if choice in corr:
            return choice in yes


def _find_bigflow_req(reqs):
    pat = re.compile(r"^bigflow\W")
    return next(filter(pat.match, reqs), None)


def maybe_upgrade_pyproject_bigflow_version(root: Path):
    pptf = root / "pyproject.toml"
    reqsf = root / "resources" / "requirements.txt"

    if not pptf.exists() or not reqsf.exists():
        return

    reqs_version =_find_bigflow_req(bigflow.build.pip.read_requirements(reqsf, False))
    reqs_version = re.sub(r"\[.*\]", "", reqs_version)  # strip extras

    ppt = toml.load(pptf)
    ppt_requires = ppt.get('build-system', {}).get('requires', [])
    ppt_version = _find_bigflow_req(ppt_requires)

    if reqs_version == ppt_version:
        logger.debug("Same bigflow version in pyproject.toml & requirements.txt")
        return

    if not reqs_version:
        logger.debug("No bigflow versin in requirements.txt")
        return

    print("Mismathed versions of bigflow")
    print(" - pyproject.toml:", ppt_version)
    print(" - requirements.txt:", reqs_version)
    print("Do you want to update version in your `pyproject.toml`?")

    if not _yes_or_no():
        return

    ppt_requires[ppt_requires.index(ppt_version)] = reqs_version
    ppt.setdefault('build-system', {})['requires'] = ppt_requires

    logger.info("Update `pyproject.toml`")
    pptf.write_text(toml.dumps(ppt))


def need_migrate_to_11(root: Path):
    return (root / "project_setup.py").exists() and not all([
        (root / "setup.py").exists(),
        (root / "pyproject.toml").exists(),
    ])


def _rename_projectsetup_to_setup(root: Path):
    prj_setup_py = root / "project_setup.py"
    setup_py = root / "setup.py"

    if prj_setup_py.exists() and setup_py.exists():
        logger.error("Both files `setup.py` and `project_setup.py` exists! File `setup_project.py` renamed to `setup.py`")
        setup_py.rename(root / "setup.py.back")
        prj_setup_py.rename(setup_py)

    elif prj_setup_py.exists():
        logger.warning("File `project_setup.py` renamed to `setup.py`")
        prj_setup_py.rename(setup_py)


def migrate__v1_0__v1_1(root: Path):

    print("Project uses old structure - it needs to be migrated in order to use bigflow >= 1.1")
    print("Project config 'project_setup.py' will be renamed into 'setup.py'")
    print("A few additional files will be created: pyproject.toml")
    print("Do you wan to run migration now?")

    if not _yes_or_no():
        print("Sorry, but you need to downgrade bigflow by running 'pip install bigflow~=1.0'")
        sys.exit(1)

    _rename_projectsetup_to_setup(root)
    project_opts = bigflow.build.dev.read_setuppy_args(root / "setup.py")
    project_name = project_opts['name']

    # Write 'pyproject.toml'
    from bigflow.scaffold import scaffold
    scaffold.migrate_project_from_10(root, project_name)

    # Check 'setup.py' is not ignored
    gitignore = (root / ".gitignore")
    if gitignore.exists():
        gitignore_content = gitignore.read_text()
        gitignore_content = "\n".join(
            x
            for x in gitignore_content.splitlines()
            if x != "setup.py"
            and x != "/setup.py"
        )
        gitignore.write_text(gitignore_content)


def check_migrate(root: Optional[Path] = None):
    # TODO: Implement proper migration detection based on `bigflow.__version__` and `requirements.txt`
    # Consider adding 'bigflow' version as argument to `bigflow.build.setup(...)`
    if root is None:
        root = Path.cwd()

    if need_migrate_to_11(root):
        logging.debug("Migrate project to 1.1")
        migrate__v1_0__v1_1(root)



import argparse
import importlib
import os
import pathlib
import subprocess
import sys
import logging

import importlib.util

from argparse import Namespace
from datetime import datetime
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import Tuple, Iterator
from typing import Optional
from glob import glob1

import bigflow as bf
import bigflow.build.pip
import bigflow.resources
import bigflow.commons as bf_commons
import bigflow.build.dist
import bigflow.build.dev
import bigflow.build.operate
import bigflow.build.spec
import bigflow.migrate

from bigflow import Config
from bigflow.deploy import deploy_dags_folder, deploy_docker_image
from bigflow.scaffold import start_project
from bigflow.version import get_version, release


logger = logging.getLogger(__name__)


def walk_module_files(root_package: Path) -> Iterator[Tuple[str, str]]:
    """
    Returning all the Python files in the `root_package`

    Example:
    walk_module_files(Path("fld")) -> [("/path/to/fld", "file1"), ("/path/to/fld", "file2")]

    @return: (absolute_path: str, name: str)
    """
    logger.debug("walk module files %r", root_package)
    for subdir, dirs, files in os.walk(str(root_package)):
        for file in files:
            if file.endswith('.py'):
                logger.debug("found python file %s/%s", subdir, file)
                yield subdir, file


def _removesuffix(s, suffix):
    return s[:-len(suffix)] if s.endswith(suffix) else s


def build_module_path(root_package: Path, module_dir: Path, module_file: str) -> str:
    """
    Returns module path that can be imported using `import_module`
    """
    full_module_file_path = str(module_dir.absolute() / module_file)
    full_module_file_path = full_module_file_path.replace(str(root_package.parent.absolute()), '')

    res = full_module_file_path
    res = _removesuffix(res, ".py")
    res = _removesuffix(res, "/__init__")
    res = res.lstrip("/").replace(os.sep, ".")
    return res


def walk_module_paths(root_package: Path) -> Iterator[str]:
    """
    Returning all the module paths in the `root_package`
    """
    logger.debug("walk module paths, root %r", root_package)
    for module_dir, module_file in walk_module_files(root_package):
        mpath = build_module_path(root_package, Path(module_dir), module_file)
        logger.debug("%s / %s / %s resolved to module %r", root_package, module_dir, module_file, mpath)
        logger.debug("path %r", mpath)
        yield mpath


def walk_modules(root_package: Path) -> Iterator[ModuleType]:
    """
    Imports all the modules in the path and returns
    """
    logger.debug("walk modules, root %r", root_package)
    for module_path in walk_module_paths(root_package):
        try:
            logger.debug("import module %r", module_path)
            logger.debug("%r", sys.path)
            yield import_module(module_path)
        except ValueError as e:
            print(f"Skipping module {module_path}. Can't import due to exception {str(e)}.")


def walk_module_objects(module: ModuleType, expect_type: type) -> Iterator[Tuple[str, type]]:
    """
    Returns module items of the set type
    """
    logger.debug("scan module %r for object of type %r", module, expect_type)
    for name, obj in module.__dict__.items():
        if isinstance(obj, expect_type):
            yield name, obj


def walk_workflows(root_package: Path) -> Iterator[bf.Workflow]:
    """
    Imports modules in the `root_package` and returns all the elements of the type bf.Workflow
    """
    logger.debug("walk workflows, root %s", root_package)
    for module in walk_modules(root_package):
        for name, workflow in walk_module_objects(module, bf.Workflow):
            yield workflow


def find_workflow(root_package: Path, workflow_id: str) -> bf.Workflow:
    """
    Imports modules and finds the workflow with id workflow_id
    """
    logger.debug("find workflow, root %s, workflow_id %r", root_package, workflow_id)
    for workflow in walk_workflows(root_package):
        if workflow.workflow_id == workflow_id:
            return workflow
    raise ValueError('Workflow with id {} not found in package {}'.format(workflow_id, root_package))

def set_configuration_env(env):
    """
    Sets 'bf_env' env variable
    """
    if env is not None:
        os.environ['bf_env'] = env
        print(f"bf_env is : {os.environ.get('bf_env', None)}")


def _init_workflow_log(workflow: bf.Workflow):
    if not workflow.log_config:
        return

    try:
        import bigflow.log
    except ImportError:
        # `log` extras is not installed?
        pass
    else:
        bigflow.log.init_workflow_logging(workflow)


def execute_job(root_package: Path, workflow_id: str, job_id: str, runtime=None):
    """
    Executes the job with the `workflow_id`, with job id `job_id`

    @param runtime: str determine partition that will be used for write operations.
    """
    w = find_workflow(root_package, workflow_id)
    _init_workflow_log(w)
    w.run_job(job_id, runtime)


def execute_workflow(root_package: Path, workflow_id: str, runtime=None):
    """
    Executes the workflow with the `workflow_id`

    @param runtime: str determine partition that will be used for write operations.
    """
    w = find_workflow(root_package, workflow_id)
    _init_workflow_log(w)
    w.run(runtime)


def read_project_name_from_setup() -> Optional[str]:
    logger.debug("Read project name from project spec")
    try:
        return bigflow.build.spec.get_project_spec().name
    except Exception as e:
        logger.warning("Unable to read project name: %s", e)
        return None


def find_root_package(project_name: Optional[str], project_dir: Optional[str]) -> Path:
    """
    Finds project package path. Tries first to find location in project_setup.PROJECT_NAME,
    and if not found then by making a path to the `root` module

    @param project_dir: Path to the root package of a project, used only when PROJECT_NAME not set
    @return: Path
    """
    if project_name is not None:
        project_name = project_name.replace("-", "_")
        return Path(project_name)
    else:
        root_module = import_module(project_dir)
        return Path(root_module.__file__.replace('__init__.py', ''))


def import_deployment_config(deployment_config_path: str, property_name: str):
    if not Path(deployment_config_path).exists():
        raise ValueError(f"Can't find deployment_config.py at '{deployment_config_path}'. "
                         f"Property '{property_name}' can't be resolved. "
                          "If your deployment_config.py is elswhere, "
                          "you can set path to it using --deployment-config-path. If you are not using deployment_config.py -- "
                         f"set '{property_name}' property as a command line argument.")
    spec = importlib.util.spec_from_file_location('deployment_config', deployment_config_path)

    if not spec:
        raise ValueError(f'Failed to load deployment_config from {deployment_config_path}. '
        'Create a proper deployment_config.py file'
        'or set all the properties via command line arguments.')

    deployment_config_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(deployment_config_module)

    if not isinstance(deployment_config_module.deployment_config, Config):
        raise ValueError('deployment_config attribute in deployment_config.py should be instance of bigflow.Config')

    return deployment_config_module.deployment_config


def cli_run(project_package: str,
            runtime: Optional[str] = None,
            full_job_id: Optional[str] = None,
            workflow_id: Optional[str] = None) -> None:
    """
    Runs the specified job or workflow

    @param project_package: str The main package of a user's project
    @param runtime: Optional[str] Date of XXX in format "%Y-%m-%d %H:%M:%S"
    @param full_job_id: Optional[str] Represents both workflow_id and job_id in a string in format "<workflow_id>.<job_id>"
    @param workflow_id: Optional[str] The id of the workflow that should be executed
    @return:
    """

    # TODO: Check that installed libs in sync with `requirements.txt`
    bigflow.build.pip.check_requirements_needs_recompile(Path("resources/requirements.txt"))

    if full_job_id is not None:
        try:
            workflow_id, job_id = full_job_id.split('.')
        except ValueError:
            raise ValueError(
                'You should specify job using the workflow_id and job_id parameters - --job <workflow_id>.<job_id>.')
        execute_job(project_package, workflow_id, job_id, runtime=runtime)
    elif workflow_id is not None:
        execute_workflow(project_package, workflow_id, runtime=runtime)
    else:
        raise ValueError('You must provide the --job or --workflow for the run command.')


def _parse_args(project_name: Optional[str], args) -> Namespace:
    parser = argparse.ArgumentParser(description=f'Welcome to BigFlow CLI.'
                                                  '\nType: bigflow {command} -h to print detailed help for a selected command.')
    parser.add_argument(
        "-v",
        "--verbose",
        action='store_true',
        default=False,
        help="Print verbose output for debugging",
    )

    subparsers = parser.add_subparsers(dest='operation',
                                       required=True,
                                       help='BigFlow command to execute')

    _create_run_parser(subparsers, project_name)
    _create_deploy_dags_parser(subparsers)
    _create_deploy_image_parser(subparsers)
    _create_deploy_parser(subparsers)

    _create_build_dags_parser(subparsers)
    _create_build_image_parser(subparsers)
    _create_build_package_parser(subparsers)
    _create_build_parser(subparsers)

    _create_project_version_parser(subparsers)
    _create_release_parser(subparsers)
    _create_start_project_parser(subparsers)
    _create_logs_parser(subparsers)

    _create_build_requirements_parser(subparsers)

    _create_codegen_parser(subparsers)

    return parser.parse_args(args)


def _create_logs_parser(subparsers):
    subparsers.add_parser('logs', description='Returns a link leading to a workflow logs in GCP Logging.')


def _create_start_project_parser(subparsers):
    subparsers.add_parser('start-project', description='Creates a scaffolding project in a current directory.')


def _create_build_parser(subparsers):
    parser = subparsers.add_parser('build', description='Builds a Docker image, DAG files and .whl package from local sources.')
    _add_build_dags_parser_arguments(parser)


def _create_build_package_parser(subparsers):
    subparsers.add_parser('build-package', description='Builds .whl package from local sources.')


def _valid_datetime(dt):
    if dt == 'NOW':
        return

    try:
        datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")
        return dt
    except ValueError:
        try:
            datetime.strptime(dt, "%Y-%m-%d")
            return dt
        except ValueError:
            raise ValueError("Not a valid date: '{0}'.".format(dt))


def _add_build_dags_parser_arguments(parser):
    parser.add_argument('-w', '--workflow',
                        type=str,
                        help="Skip or set to ALL to build DAGs from all workflows. "
                             "Set a workflow Id to build selected workflow only. "
                             "For example to build only this workflow: bigflow.Workflow(workflow_id='workflow1',"
                             " definition=[ExampleJob('job1')]) you should use --workflow workflow1")
    parser.add_argument('-t', '--start-time',
                        help='The first runtime of a workflow. '
                             'For workflows triggered hourly -- datetime in format: Y-m-d H:M:S, for example 2020-01-01 00:00:00. '
                             'For workflows triggered daily -- date in format: Y-m-d, for example 2020-01-01. '
                             'If empty or set as NOW, current hour is used.',
                        type=_valid_datetime)


def _create_build_dags_parser(subparsers):
    parser = subparsers.add_parser('build-dags',
                                   description='Builds DAG files from local sources to {current_dir}/.dags')
    _add_build_dags_parser_arguments(parser)


def _create_build_image_parser(subparsers):
    subparsers.add_parser('build-image',
                          description='Builds a docker image from local files.')


def _create_run_parser(subparsers, project_name):
    parser = subparsers.add_parser('run',
                                   description='BigFlow CLI run command -- run a workflow or job')

    group = parser.add_mutually_exclusive_group()
    group.required = True
    group.add_argument('-j', '--job',
                       type=str,
                       help='The job to start, identified by workflow id and job id in format "<workflow_id>.<job_id>".')
    group.add_argument('-w', '--workflow',
                       type=str,
                       help='The id of the workflow to start.')
    parser.add_argument('-r', '--runtime',
                        type=str, default=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        help='The date and time when this job or workflow should be started. '
                             'The default is now (%(default)s). '
                             'Examples: 2019-01-01, 2020-01-01 01:00:00')
    _add_parsers_common_arguments(parser)

    if project_name is None:
        parser.add_argument('--project-package',
                            required=True,
                            type=str,
                            help='The main package of your project. '
                                 'Should contain `setup.py`')


def _add_parsers_common_arguments(parser):
    parser.add_argument('-c', '--config',
                        type=str,
                        help='Config environment name that should be used. For example: dev, prod.'
                             ' If not set, default Config name will be used.'
                             ' This env name is applied to all bigflow.Config objects that are defined by'
                             ' individual workflows as well as to deployment_config.py.')


def _add_deploy_parsers_common_arguments(parser):
    parser.add_argument('-a', '--auth-method',
                        type=str,
                        default='local_account',
                        help='One of two authentication method: '
                             'local_account -- you are using credentials of your local user authenticated in gcloud; '
                             'vault -- credentials for service account are obtained from Vault. '
                             'Default: local_account',
                        choices=['local_account', 'vault'])
    parser.add_argument('-ve', '--vault-endpoint',
                        type=str,
                        help='URL of a Vault endpoint to get OAuth token for service account. '
                             'Required if auth-method is vault. '
                             'If not set, will be read from deployment_config.py.'
                        )
    parser.add_argument('-vs', '--vault-secret',
                        type=str,
                        help='Vault secret token. '
                             'Required if auth-method is vault.'
                        )
    parser.add_argument('-dc', '--deployment-config-path',
                        type=str,
                        help='Path to the deployment_config.py file. '
                             'If not set, {current_dir}/deployment_config.py will be used.')

    _add_parsers_common_arguments(parser)


def _create_deploy_parser(subparsers):
    parser = subparsers.add_parser('deploy',
                                   description='Performs complete deployment. Uploads DAG files from local DAGs folder '
                                               'to Composer and uploads Docker image to Container Registry.')

    _add_deploy_dags_parser_arguments(parser)
    _add_deploy_image_parser_arguments(parser)
    _add_deploy_parsers_common_arguments(parser)


def _create_deploy_image_parser(subparsers):
    parser = subparsers.add_parser('deploy-image',
                                   description='Uploads Docker image to Container Registry.'
                                   )

    _add_deploy_image_parser_arguments(parser)
    _add_deploy_parsers_common_arguments(parser)


def _create_deploy_dags_parser(subparsers):
    parser = subparsers.add_parser('deploy-dags',
                                   description='Uploads DAG files from local DAGs folder to Composer.')

    _add_deploy_dags_parser_arguments(parser)
    _add_deploy_parsers_common_arguments(parser)


def _create_project_version_parser(subparsers):
    subparsers.add_parser('project-version', aliases=['pv'], description='Prints project version')


def _create_release_parser(subparsers):
    parser = subparsers.add_parser('release', description='Creates a new release tag')
    parser.add_argument('-i', '--ssh-identity-file',
                        type=str,
                        help="Path to the identity file, used to authorize push to remote repository"
                             " If not specified, default ssh configuration will be used.")


def _add_deploy_image_parser_arguments(parser):
    parser.add_argument('-i', '--image-tar-path',
                        type=str,
                        help='Path to a Docker image file. The file name must contain version number with the following naming schema: image-{version}.tar')
    parser.add_argument('-r', '--docker-repository',
                        type=str,
                        help='Name of a local and target Docker repository. Typically, a target repository is hosted by Google Cloud Container Registry.'
                             ' If so, with the following naming schema: {HOSTNAME}/{PROJECT-ID}/{IMAGE}.'
                        )

def _add_deploy_dags_parser_arguments(parser):
    parser.add_argument('-dd', '--dags-dir',
                        type=str,
                        help="Path to the folder with DAGs to deploy."
                             " If not set, {current_dir}/.dags will be used.")
    parser.add_argument('-cdf', '--clear-dags-folder',
                        action='store_true',
                        help="Clears the DAGs bucket before uploading fresh DAG files. "
                             "Default: False")

    parser.add_argument('-p', '--gcp-project-id',
                        help="Name of your Google Cloud Platform project."
                             " If not set, will be read from deployment_config.py")

    parser.add_argument('-b', '--dags-bucket',
                        help="Name of the target Google Cloud Storage bucket which underlies DAGs folder of your Composer."
                             " If not set, will be read from deployment_config.py")


def read_project_package(args):
    return args.project_package if hasattr(args, 'project_package') else None


def _resolve_deployment_config_path(args):
    if args.deployment_config_path:
        return args.deployment_config_path
    return os.path.join(os.getcwd(), 'deployment_config.py')


def _resolve_dags_dir(args):
    if args.dags_dir:
        return args.dags_dir
    return os.path.join(os.getcwd(), '.dags')


def _resolve_vault_endpoint(args):
    if args.auth_method == 'vault':
        return _resolve_property(args, 'vault_endpoint')
    else:
        return None


def _resolve_property(args, property_name):
    cli_atr = getattr(args, property_name)
    if cli_atr:
        return cli_atr
    else:
        config = import_deployment_config(_resolve_deployment_config_path(args), property_name)
        return config.resolve_property(property_name, args.config)


def _cli_deploy_dags(args):
    try:
        vault_secret = _resolve_property(args, 'vault_secret')
    except ValueError:
        vault_secret = None
    deploy_dags_folder(dags_dir=_resolve_dags_dir(args),
                       dags_bucket=_resolve_property(args, 'dags_bucket'),
                       clear_dags_folder=args.clear_dags_folder,
                       auth_method=args.auth_method,
                       vault_endpoint=_resolve_vault_endpoint(args),
                       vault_secret=vault_secret,
                       project_id=_resolve_property(args, 'gcp_project_id')
                       )


def _load_image_from_tar(image_tar_path: str):
    print(f'Loading Docker image from {image_tar_path} ...', )


def _cli_deploy_image(args):
    docker_repository = _resolve_property(args, 'docker_repository')
    try:
        vault_secret = _resolve_property(args, 'vault_secret')
    except ValueError:
        vault_secret = None
    image_tar_path = args.image_tar_path if args.image_tar_path  else find_image_file()

    deploy_docker_image(image_tar_path=image_tar_path,
                        auth_method=args.auth_method,
                        docker_repository=docker_repository,
                        vault_endpoint=_resolve_vault_endpoint(args),
                        vault_secret=vault_secret)


def find_image_file():
    # TODO parametrize ".image" using settings from build.py
    files = glob1(".image", "*-*.tar")
    if files:
        return os.path.join(".image", files[0])
    else:
        raise ValueError('File containing image to deploy not found')


def _cli_build_image(args):
    prj = bigflow.build.spec.get_project_spec()
    bigflow.build.operate.build_image(prj)


def _cli_build_package():
    prj = bigflow.build.spec.get_project_spec()
    bigflow.build.operate.build_package(prj)


def _cli_build_dags(args):
    prj = bigflow.build.spec.get_project_spec()
    bigflow.build.operate.build_dags(
        prj,
        start_time=args.start_time if _is_starttime_selected(args) else datetime.now().strftime("%Y-%m-%d %H:00:00"),
        workflow_id=args.workflow if _is_workflow_selected(args) else None,
    )


def _cli_build(args):
    prj = bigflow.build.spec.get_project_spec()
    bigflow.build.operate.build_project(
        prj,
        start_time=args.start_time if _is_starttime_selected(args) else datetime.now().strftime("%Y-%m-%d %H:00:00"),
        workflow_id=args.workflow if _is_workflow_selected(args) else None,
    )


def _create_build_requirements_parser(subparsers):
    parser = subparsers.add_parser(
        'build-requirements',
        description="Compiles requirements.txt from *.in specs",
    )
    parser.add_argument(
        'in_file',
        type=str,
        nargs='?',
        default="resources/requirements.in",  # FIXME: read 'project_setup.py'
    )


def _create_codegen_parser(subparsers: argparse._SubParsersAction):
    parser = subparsers.add_parser('codegen', description="Various codegeneration tools")
    ss = parser.add_subparsers()

    p = ss.add_parser('pin-dataflow-requirements')
    p.set_defaults(func=_cli_codegen_pin_dataflow_requirements)


def _cli_build_requirements(args):
    in_file = pathlib.Path(args.in_file)
    bigflow.build.pip.pip_compile(in_file)


def _cli_codegen(args):
    args.func(args)


def _cli_codegen_pin_dataflow_requirements(args):
    import bigflow.build.dataflow.dependency_checker as dc
    dc.sync_requirements_with_dataflow_workers()


def _is_workflow_selected(args):
    return args.workflow and args.workflow != 'ALL'


def _is_starttime_selected(args):
    return args.start_time and args.start_time != 'NOW'


def project_type_input():
    project_type = input("Would you like to create basic or advanced project? Default basic. Type 'a' for advanced.\n")
    return project_type if project_type else 'b'


def project_number_input():
    project_number = input('How many GCP projects would you like to use? '
                           'It allows to deploy your workflows to more than one project. Default 2\n')
    return project_number if project_number else '2'


def gcloud_project_list():
    return subprocess.getoutput('gcloud projects list')


def get_default_project_from_gcloud():
    return subprocess.getoutput('gcloud config get-value project')


def project_id_input(n):
    if n == 0:
        project = input(f'Enter a GCP project ID that you are going to use in your BigFlow project. '
                        f'Choose a project from the list above. '
                        f'If not provided default project: {get_default_project_from_gcloud()} will be used.\n')
    else:
        project = input(f'Enter a #{n} GCP project ID that you are going to use in your BigFlow project. '
                        f'Choose a project from the list above.'
                        f' If not provided default project: {get_default_project_from_gcloud()} will be used.\n')
    return project


def gcp_project_flow(n):
    projects_list = gcloud_project_list()
    print(projects_list)
    return gcp_project_input(n, projects_list)


def gcp_project_input(n, projects_list):
    project = project_id_input(n)
    if project == '':
        return get_default_project_from_gcloud()
    if project not in projects_list:
        print(f'You do not have access to {project}. Try another project from the list.\n')
        return gcp_project_input(n, projects_list)
    return project


def gcp_bucket_input():
    return input('Enter a Cloud Composer Bucket name where DAG files will be stored.\n')


def environment_name_input(envs):
    environment_name = input('Enter an environment name. Default dev\n')
    if environment_name in envs:
        print(f'Environment with name{environment_name} is already defined. Try another name.\n')
        return environment_name_input(envs)
    return environment_name if environment_name else 'dev'


def project_name_input():
    return input('Enter the project name. It should be valid python package name. '
                 'It will be used as a main directory of your project and bucket name used by dataflow to run jobs.\n')


def _cli_start_project():
    config = {'is_basic': False, 'project_name': project_name_input(), 'projects_id': [], 'composers_bucket': [], 'envs': []}
    if False:
        for n in range(0, int(project_number_input())):
            config['projects_id'].append(gcp_project_flow(n))
            config['composers_bucket'].append(gcp_bucket_input())
            config['envs'].append(environment_name_input(config['envs']))
    else:
        config['is_basic'] = True
        config['projects_id'].append(gcp_project_flow(0))
        config['composers_bucket'].append(gcp_bucket_input())

        config['pyspark_job'] = True

    start_project(**config)
    print('Bigflow project created successfully.')


def _cli_project_version(args):
    print(get_version())


def _cli_release(args):
    release(args.ssh_identity_file)


def init_console_logging(verbose):
    if verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(asctime)s| %(message)s",
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format="%(message)s",
        )


def cli_logs(root_package):
    import bigflow.log as log

    projects_id = []
    workflows_links = {}
    for workflow in walk_workflows(root_package):
        if workflow.log_config:
            projects_id.append((workflow.log_config['gcp_project_id'], workflow.workflow_id))
            workflows_links[workflow.workflow_id] = log.workflow_logs_link_for_cli(workflow.log_config, workflow.workflow_id)
    if not projects_id:
        raise Exception("Found no workflows with configured logging.")
    deduplicated_projects_id = sorted(set(projects_id), key=lambda x: projects_id.index(x))
    infra_links = log.infrastructure_logs_link_for_cli(deduplicated_projects_id)
    log.print_log_links_message(workflows_links, infra_links)


def _is_log_module_installed():
    try:
        import bigflow.log
        return True
    except ImportError:
        raise Exception("bigflow.log module not found. You need install bigflow with 'log' extras.")


def cli(raw_args) -> None:
    bigflow.build.dev.install_syspath()
    bigflow.migrate.check_migrate()

    project_name = read_project_name_from_setup()
    parsed_args = _parse_args(project_name, raw_args)
    init_console_logging(parsed_args.verbose)

    operation: str = parsed_args.operation

    if operation == 'run':
        set_configuration_env(parsed_args.config)
        root_package = find_root_package(project_name, read_project_package(parsed_args))
        cli_run(root_package, parsed_args.runtime, parsed_args.job, parsed_args.workflow)
    elif operation == 'deploy-image':
        _cli_deploy_image(parsed_args)
    elif operation == 'deploy-dags':
        _cli_deploy_dags(parsed_args)
    elif operation == 'deploy':
        _cli_deploy_image(parsed_args)
        _cli_deploy_dags(parsed_args)
    elif operation == 'build-dags':
        _cli_build_dags(parsed_args)
    elif operation == 'build-image':
        _cli_build_image(parsed_args)
    elif operation == 'build-package':
        _cli_build_package()
    elif operation == 'build':
        _cli_build(parsed_args)
    elif operation == 'start-project':
        _cli_start_project()
    elif operation == 'project-version' or operation == 'pv':
        _cli_project_version(parsed_args)
    elif operation == 'release':
        _cli_release(parsed_args)
    elif operation == 'logs':
        _is_log_module_installed()
        root_package = find_root_package(project_name, None)
        cli_logs(root_package)
    elif operation == 'build-requirements':
        _cli_build_requirements(parsed_args)
    elif operation == 'codegen':
        _cli_codegen(parsed_args)
    else:
        raise ValueError(f'Operation unknown - {operation}')



"""Actual implementaion of buid/distribution operations"""

import os
import subprocess
import shutil
import typing
import logging
import typing

from pathlib import Path

import bigflow.resources
import bigflow.dagbuilder
import bigflow.version
import bigflow.build.pip
import bigflow.build.dev
import bigflow.build.dist
import bigflow.build.dataflow.dependency_checker
import bigflow.commons as bf_commons

from bigflow.build.spec import BigflowProjectSpec


logger = logging.getLogger(__name__)


def run_tests(project_spec: BigflowProjectSpec):
    logger.info('Runing tests...')
    output_dir = "build/junit-reports"
    try:
        bf_commons.run_process([
            "python", "-m", "xmlrunner", "discover",
            "-s", ".",
            "-t", project_spec.project_dir,
            "-o", output_dir,
        ])
    except subprocess.CalledProcessError:
        logger.error("Test suite was FAILED")
        exit(1)

    logger.info("Test suite was PASSED")


def _export_docker_image_to_file(tag: str, target_dir: Path, version: str):
    image_target_path = target_dir / f"image-{version}.tar"
    logger.info("Exporting the image to %s ...", image_target_path)
    bf_commons.run_process(["docker", "image", "save", "-o", image_target_path, bf_commons.get_docker_image_id(tag)])


def _build_docker_image(project_dir: Path, tag: str):
    logger.debug("Run docker build...")
    bf_commons.run_process(f'docker build {project_dir} --tag {tag}')


def build_image(
    project_spec: BigflowProjectSpec,
):
    logger.info("Building docker image...")
    clear_image_leftovers(project_spec)

    image_dir = project_spec.project_dir / ".image"
    os.mkdir(image_dir)

    tag = bf_commons.build_docker_image_tag(project_spec.docker_repository, project_spec.version)
    _build_docker_image(project_spec.project_dir, tag)

    try:
        _export_docker_image_to_file(tag, image_dir, project_spec.version)
        dconf_file = Path(project_spec.deployment_config_file)
        shutil.copyfile(dconf_file, image_dir / dconf_file.name)
    finally:
        bf_commons.remove_docker_image_from_local_registry(tag)

    logger.info("Docker image was built")


def build_dags(
    project_spec: BigflowProjectSpec,
    start_time: str,
    workflow_id: typing.Optional[str] = None,
):
    logger.info("Building airflow DAGs...")
    clear_dags_leftovers(project_spec)

    # TODO: Move common frunctions from bigflow.cli to bigflow.commons (or other shared module)
    from bigflow.cli import _valid_datetime, walk_workflows
    _valid_datetime(start_time)

    cnt = 0
    for root_package in project_spec.packages:
        if "." in root_package:
            # leaf package
            continue

        for workflow in walk_workflows(project_spec.project_dir / root_package):
            if workflow_id is not None and workflow_id != workflow.workflow_id:
                continue
            logger.info("Generating DAG file for %s", workflow.workflow_id)
            cnt += 1
            bigflow.dagbuilder.generate_dag_file(
                str(project_spec.project_dir),
                project_spec.docker_repository,
                workflow,
                start_time,
                project_spec.version,
                root_package,
            )

    logger.info("Geneated %d DAG files", cnt)


def _rmtree(p: Path):
    logger.info("Removing directory %s", p)
    shutil.rmtree(p, ignore_errors=True)


def clear_image_leftovers(project_spec: BigflowProjectSpec):
    _rmtree(project_spec.project_dir / ".image")


def clear_dags_leftovers(project_spec: BigflowProjectSpec):
    _rmtree(project_spec.project_dir / ".dags")


def build_package(project_spec: BigflowProjectSpec):
    logger.info('Building python package')
    clear_package_leftovers(project_spec)
    run_tests(project_spec)
    bigflow.build.dist.run_setup_command(project_spec, 'bdist_wheel')


def clear_package_leftovers(project_spec: BigflowProjectSpec):
    _rmtree(project_spec.project_dir / "build")
    _rmtree(project_spec.project_dir / "dist")
    _rmtree(project_spec.project_dir / f"{project_spec.name}.egg")


def build_project(
    project_spec: BigflowProjectSpec,
    start_time: str,
    workflow_id: typing.Optional[str] = None,
):
    logger.info("Build the project")
    build_package(project_spec)
    build_image(project_spec)
    build_dags(project_spec, start_time, workflow_id=workflow_id)
    logger.info("Project was built")
